# -*- coding: utf-8 -*-
"""
admin security authenticators module.
"""

from pyrin.security.enumerations import InternalAuthenticatorEnum
from pyrin.security.authentication.decorators import authenticator
from pyrin.security.authentication.handlers.internal import InternalTokenAuthenticator


@authenticator()
class AdminTokenAuthenticator(InternalTokenAuthenticator):
    """
    internal token authenticator class.
    """

    REFRESH_TOKEN_HOLDER = 'Admin-Refresh-Auth'
    _name = InternalAuthenticatorEnum.ADMIN

    def _get_extra_user_info(self, user, **options):
        """
        gets extra user info for given internal user if required.

        :param ROW_RESULT user: internal user to get its extra info.

        :rtype: dict
        """

        return dict(admin_access=user.admin_access)



# -*- coding: utf-8 -*-
"""
authentication hooks module.
"""

import pyrin.security.authentication.services as authentication_services

from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def after_packages_loaded(self):
        """
        this method will be called after all application packages have been loaded.
        """

        authentication_services.validate_authenticators()



# -*- coding: utf-8 -*-
"""
validator auto decorators module.
"""

import pyrin.validator.auto.services as auto_validator_services


def auto_validator_hook():
    """
    decorator to register an auto validator hook.

    :raises InvalidAutoValidatorHookTypeError: invalid auto validator hook type error.

    :returns: auto validator hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available auto validator hooks.

        :param type cls: auto validator hook class.

        :returns: auto validator hook class.
        :rtype: type
        """

        instance = cls()
        auto_validator_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
users mixin module.
"""

from sqlalchemy import Unicode
from sqlalchemy.ext.hybrid import hybrid_property

from pyrin.core.globals import _
from pyrin.core.enumerations import CoreEnum, EnumMember
from pyrin.database.orm.sql.schema.columns import AutoPKColumn, StringColumn, \
    SmallIntegerColumn, BooleanColumn, TimeStampColumn, HiddenColumn


class UserPKMixin:
    """
    user pk mixin class.
    """

    id = AutoPKColumn(name='id')


class UserCredentialMixin:
    """
    user credential mixin class.
    """

    username = StringColumn(name='username', min_length=1, max_length=30,
                            unique=True, index=True, nullable=False)
    password_hash = HiddenColumn(name='password_hash', type_=Unicode(250), nullable=False)


class UserCommonMixin:
    """
    user common mixin class.
    """

    class GenderEnum(CoreEnum):
        FEMALE = EnumMember(0, _('Female'))
        MALE = EnumMember(1, _('Male'))
        OTHER = EnumMember(2, _('Other'))

    first_name = StringColumn(name='first_name', min_length=1, max_length=50, nullable=False)
    last_name = StringColumn(name='last_name', min_length=1, max_length=50, nullable=False)
    mobile = StringColumn(name='mobile', min_length=1, max_length=25, unique=True)
    email = StringColumn(name='email', min_length=6, max_length=50, unique=True)
    gender = SmallIntegerColumn(name='gender', check_in=GenderEnum)
    is_active = BooleanColumn(name='is_active', nullable=False, default=True)
    last_login_at = TimeStampColumn(name='last_login_at', allow_write=False)

    @hybrid_property
    def fullname(self):
        """
        gets the full name of this internal user at instance level.

        :rtype: str
        """

        return f'{self.first_name} {self.last_name}'

    @fullname.expression
    def fullname(cls):
        """
        gets the full name of this internal user at expression level.

        :rtype: str
        """

        return cls.first_name + ' ' + cls.last_name


class UserAccessMixin:
    """
    user access mixin class.
    """

    is_superuser = BooleanColumn(name='is_superuser', nullable=False, default=False)
    admin_access = BooleanColumn(name='admin_access', nullable=False, default=False)
    audit_access = BooleanColumn(name='audit_access', nullable=False, default=False)
    swagger_access = BooleanColumn(name='swagger_access', nullable=False, default=False)



# -*- coding: utf-8 -*-
"""
users internal exceptions module.
"""

from pyrin.validator.exceptions import ValidationError
from pyrin.core.exceptions import CoreBusinessException, CoreException


class InternalUsersManagerException(CoreException):
    """
    internal users manager exception.
    """
    pass


class InternalUsersManagerBusinessException(CoreBusinessException,
                                            InternalUsersManagerException):
    """
    internal users manager business exception.
    """
    pass


class InternalUserNotFoundError(InternalUsersManagerBusinessException):
    """
    internal user not found error.
    """
    pass


class PasswordsDoNotMatchError(ValidationError,
                               InternalUsersManagerBusinessException):
    """
    passwords do not match error.
    """
    pass



# -*- coding: utf-8 -*-
"""
audit security exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.security.exceptions import AuthorizationFailedError


class AuditSecurityException(CoreException):
    """
    audit security exception.
    """
    pass


class AuditSecurityBusinessException(CoreBusinessException,
                                     AuditSecurityException):
    """
    audit security business exception.
    """
    pass


class AuditAccessNotAllowedError(AuthorizationFailedError,
                                 AuditSecurityBusinessException):
    """
    audit access not allowed error.
    """
    pass



# -*- coding: utf-8 -*-
"""
authentication decorators module.
"""

import pyrin.security.authentication.services as authentication_services


def authenticator(*args, **kwargs):
    """
    decorator to register an authenticator.

    :param object args: authenticator class constructor arguments.
    :param object kwargs: authenticator class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           authenticator with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidAuthenticatorTypeError: invalid authenticator type error.
    :raises DuplicatedAuthenticatorError: duplicated authenticator error.

    :returns: authenticator class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available authenticators.

        :param type cls: authenticator class.

        :returns: authenticator class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        authentication_services.register_authenticator(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
admin page enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class TableTypeEnum(CoreEnum):
    """
    table type enum.
    """

    DENSE = 'dense'
    DEFAULT = 'default'


class PaginationTypeEnum(CoreEnum):
    """
    pagination type enum.
    """

    NORMAL = 'normal'
    STEPPED = 'stepped'


class PaginationPositionEnum(CoreEnum):
    """
    pagination position enum.
    """

    TOP = 'top'
    BOTTOM = 'bottom'
    BOTH = 'both'


class FormatEnum(CoreEnum):
    """
    format enum.
    """

    NUMERIC = 'numeric'
    TWO_DIGIT = '2-digit'


class MonthFormatEnum(CoreEnum):
    """
    month format enum.
    """

    NUMERIC = 'numeric'
    TWO_DIGIT = '2-digit'
    LONG = 'long'
    SHORT = 'short'
    NARROW = 'narrow'


class HourCycleEnum(CoreEnum):
    """
    hour cycle enum.
    """

    H23 = 'h23'
    H24 = 'h24'
    H11 = 'h11'
    H12 = 'h12'


class ButtonTypeEnum(CoreEnum):
    """
    button type enum.
    """

    CONTAINED = 'contained'
    OUTLINED = 'outlined'
    TEXT = 'text'


class LinkTypeEnum(CoreEnum):
    """
    link type enum.
    """

    BUTTON = 'button'
    LINK = 'link'



# -*- coding: utf-8 -*-
"""
admin exceptions module.
"""

from pyrin.api.router.handlers.exceptions import URLNotFoundError
from pyrin.core.exceptions import CoreException, CoreBusinessException, \
    CoreMethodNotAllowedError


class AdminManagerException(CoreException):
    """
    admin manager exception.
    """
    pass


class AdminManagerBusinessException(CoreBusinessException,
                                    AdminManagerException):
    """
    admin manager business exception.
    """
    pass


class InvalidAdminPageTypeError(AdminManagerException):
    """
    invalid admin page type error.
    """
    pass


class DuplicatedAdminPageError(AdminManagerException):
    """
    duplicated admin page error.
    """
    pass


class AdminPageNotFoundError(URLNotFoundError, AdminManagerBusinessException):
    """
    admin page not found error.
    """
    pass


class AdminOperationNotAllowedError(CoreMethodNotAllowedError, AdminManagerBusinessException):
    """
    admin operation not allowed error.
    """
    pass


class AdminPagesHaveNotLoadedError(AdminManagerException):
    """
    admin pages have not loaded error.
    """
    pass



# -*- coding: utf-8 -*-
"""
users internal component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.users.internal import InternalUsersPackage
from pyrin.users.internal.manager import InternalUsersManager


@component(InternalUsersPackage.COMPONENT_NAME)
class InternalUsersComponent(Component, InternalUsersManager):
    """
    internal users component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
filtering services module.
"""

from pyrin.application.services import get_component
from pyrin.filtering import FilteringPackage


def filter(filters, *entity, **options):
    """
    gets filtering expressions for given entity types based on given filters.

    **NOTE:**

    if a filter name is available in both `labeled_filters` and columns of
    entities, the value of `labeled_filters` will be used.

    if a filter name is available in columns of more than one entity, the
    first found entity will be used.

    :param dict filters: filters to be applied.

    :param type[pyrin.database.model.base.BaseEntity] entity: entity class type to
                                                              be used for filtering.

    :keyword list[InstrumentedAttribute | hybrid_property] ignore: columns to be ignored
                                                                   from filtering. this
                                                                   only has effect on
                                                                   `entity` and will be
                                                                   ignored for `labeled_filters`.

    :keyword bool remove: remove all keys that are applied as filter from filters input.
                          defaults to True if not provided.

    :keyword bool readable: specifies that any column or attribute
                            which has `allow_read=False` or its name
                            starts with underscore `_`, should not
                            be included in filtering.
                            defaults to True if not provided.
                            this only has effect on `entity` and will be
                            ignored for `labeled_filters`.

    :keyword dict labeled_filters: a dict containing all columns that should have
                                   a different filter name than their actual column
                                   name. for example:
                                   {'city_name': CityEntity.name,
                                    'person_name': PersonEntity.name}

    :keyword bool ignore_duplicates: specifies that if multiple labeled filters
                                     point to the same column and have the same value
                                     in provided filters, only add one of them in
                                     where clause. defaults to True if not provided.
                                     the main usage for this is in admin page which
                                     produces inclusive filters dynamically.

    :returns: list of expressions for filtering.
    :rtype: list
    """

    return get_component(FilteringPackage.COMPONENT_NAME).filter(filters, *entity, **options)


def get_range_filter_names(name):
    """
    gets range filter names for given name.

    it returns a tuple of two items. if name is `created_at`, then
    it returns (`from_created_at`, `to_created_at`) as range filters.

    :param str name: column name.

    :returns: tuple[str lower_bound_name, str upper_bound_name]
    :rtype: tuple[str, str]
    """

    return get_component(FilteringPackage.COMPONENT_NAME).get_range_filter_names(name)


def is_range_supported_type(type_):
    """
    gets a value indicating that range operations are supported for given type.

    :param type type_: type to be checked for range support.

    :rtype: bool
    """

    return get_component(FilteringPackage.COMPONENT_NAME).is_range_supported_type(type_)


def is_range_supported_column(column):
    """
    gets a value indicating that range operations are supported for given column.

    columns which their python type is not in `RANGE_SUPPORTED_TYPES`,
    do not support range operations.

    :param sqlalchemy.orm.attributes.InstrumentedAttribute column: column to be checked
                                                                   for range support.

    :rtype: bool
    """

    return get_component(FilteringPackage.COMPONENT_NAME).is_range_supported_column(column)



# -*- coding: utf-8 -*-
"""
authorization services module.
"""

import pyrin.configuration.services as config_services

from pyrin.application.services import get_component
from pyrin.caching.decorators import custom_cached
from pyrin.security.authorization import AuthorizationPackage


cache_name = config_services.get('security', 'authorization', 'cache_name')
cache_expire = config_services.get('security', 'authorization', 'cache_expire')
if cache_name is not None:
    @custom_cached(cache_name, expire=cache_expire, consider_user=True, refreshable=False)
    def authorize(user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will consider it as authorized.

        :keyword dict user_info: user info to be used for authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        return get_component(AuthorizationPackage.COMPONENT_NAME).authorize(user, permissions,
                                                                            **options)


    @custom_cached(cache_name, expire=cache_expire, consider_user=True, refreshable=False)
    def is_authorized(permissions, user=None, **options):
        """
        gets a value indicating that specified user is authorized for given permissions.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for authorization.

        :param user: user identity to be checked for authorization.
                     if not provided, current user will be used.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will return True as authorized.

        :keyword dict user_info: user info to be used for authorization.
                                 if not provided, current user info will be used.

        :rtype: bool
        """

        return get_component(AuthorizationPackage.COMPONENT_NAME).is_authorized(permissions, user,
                                                                                **options)

else:
    def authorize(user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will consider it as authorized.

        :keyword dict user_info: user info to be used for authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        return get_component(AuthorizationPackage.COMPONENT_NAME).authorize(user, permissions,
                                                                            **options)


    def is_authorized(permissions, user=None, **options):
        """
        gets a value indicating that specified user is authorized for given permissions.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for authorization.

        :param user: user identity to be checked for authorization.
                     if not provided, current user will be used.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will return True as authorized.

        :keyword dict user_info: user info to be used for authorization.
                                 if not provided, current user info will be used.

        :rtype: bool
        """

        return get_component(AuthorizationPackage.COMPONENT_NAME).is_authorized(permissions, user,
                                                                                **options)


def authorizer_exists(name):
    """
    gets a value indicating that an authorizer with given name exists.

    :param str name: authorizer name.

    :rtype: bool
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).authorizer_exists(name)


def get_authorizer(name, **options):
    """
    gets the authorizer with given name.

    :keyword str name: authorizer name.

    :raises AuthorizerNotFoundError: authorizer not found error.

    :rtype: AbstractAuthorizerBase
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).get_authorizer(name, **options)


def register_authorizer(instance, **options):
    """
    registers a new authorizer or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an authorizer which is already registered.

    :param AbstractAuthorizerBase instance: authorizer to be registered.
                                            it must be an instance of
                                            AbstractAuthorizerBase.

    :keyword bool replace: specifies that if there is another registered
                           authorizer with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidAuthorizerTypeError: invalid authorizer type error.
    :raises DuplicatedAuthorizerError: duplicated authorizer error.
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).register_authorizer(instance,
                                                                                  **options)


def validate_authorizers():
    """
    validates that all routes with permissions have their authorizers present.

    :raises AuthorizerNotFoundError: authorizer not found error.
    """

    return get_component(AuthorizationPackage.COMPONENT_NAME).validate_authorizers()



# -*- coding: utf-8 -*-
"""
authentication package.
"""

from pyrin.packaging.base import Package


class AuthenticationPackage(Package):
    """
    authentication package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'security.authentication.component'
    CONFIG_STORE_NAMES = ['authentication']
    DEPENDS = ['pyrin.validator',
               'pyrin.configuration',
               'pyrin.database.model']



# -*- coding: utf-8 -*-
"""
swagger decorators module.
"""

import pyrin.api.swagger.services as swagger_services


def tag(*args, **kwargs):
    """
    decorator to register a tag.

    :param object args: tag class constructor arguments.
    :param object kwargs: tag class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           tag with the same name, replace it with the
                           new one, otherwise raise an error. defaults to False.

    :raises InvalidTagTypeError: invalid tag type error.
    :raises DuplicatedTagError: duplicated tag error.

    :returns: tag class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available tags.

        :param type cls: tag class.

        :returns: tag class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        swagger_services.register_tag(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
security exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreBusinessException, CoreException


class SecurityManagerException(CoreException):
    """
    security manager exception.
    """
    pass


class SecurityManagerBusinessException(CoreBusinessException,
                                       SecurityManagerException):
    """
    security manager business exception.
    """
    pass


class InvalidPasswordLengthError(SecurityManagerBusinessException):
    """
    invalid password length error.
    """
    pass


class InvalidEncryptionTextLengthError(SecurityManagerBusinessException):
    """
    invalid encryption text length error.
    """
    pass


class AuthenticationFailedError(SecurityManagerBusinessException):
    """
    authentication failed error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of AuthenticationFailedError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.UNAUTHORIZED


class AuthorizationFailedError(SecurityManagerBusinessException):
    """
    authorization failed error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of AuthorizationFailedError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.FORBIDDEN


class UserIsNotActiveError(AuthorizationFailedError):
    """
    user is not active error.
    """
    pass



# -*- coding: utf-8 -*-
"""
authentication validators module.
"""

from pyrin.validator.decorators import validator
from pyrin.validator.handlers.string import StringValidator


@validator('authentication', 'username')
@validator('authentication', 'password')
class CredentialValidator(StringValidator):
    """
    credential validator class.
    """

    default_nullable = False
    default_allow_blank = False
    default_allow_whitespace = False



# -*- coding: utf-8 -*-
"""
authentication exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class AuthenticationManagerException(CoreException):
    """
    authentication manager exception.
    """
    pass


class AuthenticationManagerBusinessException(CoreBusinessException,
                                             AuthenticationManagerException):
    """
    authentication manager business exception.
    """
    pass


class InvalidAuthenticatorTypeError(AuthenticationManagerException):
    """
    invalid authenticator type error.
    """
    pass


class DuplicatedAuthenticatorError(AuthenticationManagerException):
    """
    duplicated authenticator error.
    """
    pass


class AuthenticatorNotFoundError(AuthenticationManagerException):
    """
    authenticator not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
filtering manager module.

this manager exposes an interface to perform filtering on entities.
the main usage for this type of filtering is in `find` services to reduce the relevant code.
there are some rules that need to be noticed:

1. filtering on single values (non-string) will be performed using `==`.
2. filtering on single values (string) will be performed using `icontains`.
3. filtering on list values will be performed using `in`.
4. filtering on range values will be performed using `>=` and `<=`.
"""

import re

from decimal import Decimal
from datetime import datetime, date, time

from sqlalchemy import func
from sqlalchemy.orm import InstrumentedAttribute

import pyrin.utils.misc as misc_utils
import pyrin.utils.sqlalchemy as sqlalchemy_utils

from pyrin.core.structs import Manager
from pyrin.core.globals import LIST_TYPES
from pyrin.filtering import FilteringPackage


class FilteringManager(Manager):
    """
    filtering manager class.
    """

    package_class = FilteringPackage

    # types that support range operations. for example range filtering or range validation.
    RANGE_SUPPORTED_TYPES = (str, int, float, Decimal, datetime, date, time)

    # keyword prefixes that will be used for range operations.
    # for example range filtering or range validation.
    FROM_KEYWORD_PREFIX = 'from_'
    TO_KEYWORD_PREFIX = 'to_'

    # these regexes will be used to remove prefixes from filter names.
    FROM_REGEX = re.compile(f'^{FROM_KEYWORD_PREFIX}')
    TO_REGEX = re.compile(f'^{TO_KEYWORD_PREFIX}')

    def _is_from_range_filter(self, name):
        """
        gets a value indicating that given name is a `from` range filter name.

        :param str name: filter name.

        :rtype: bool
        """

        return name.startswith(self.FROM_KEYWORD_PREFIX) and name != self.FROM_KEYWORD_PREFIX

    def _is_to_range_filter(self, name):
        """
        gets a value indicating that given name is a `to` range filter name.

        :param str name: filter name.

        :rtype: bool
        """

        return name.startswith(self.TO_KEYWORD_PREFIX) and name != self.TO_KEYWORD_PREFIX

    def _is_range_filter_name(self, name):
        """
        gets a value indicating that given name is a range filter name.

        :param str name: filter name.

        :rtype: bool
        """

        return self._is_from_range_filter(name) or self._is_to_range_filter(name)

    def _get_related_column_name(self, name):
        """
        gets the related column name for given range filter name.

        for example:
        if name='from_age' is provided, the related column name would be `age`.

        from_year => year
        to_year => year
        from_id => id
        register_date => register_date

        note that it may return the same input name as well.

        :param str name: range filter name to get its related column name.

        :rtype: str
        """

        if self._is_from_range_filter(name):
            return re.sub(self.FROM_REGEX, '', name)

        elif self._is_to_range_filter(name):
            return re.sub(self.TO_REGEX, '', name)

        return name

    def _get_related_column(self, name, *scope, **options):
        """
        gets the column with given name from the first entity that has it.

        it returns None if the column name is not available in columns
        or expression level hybrid properties of any provided entities.

        :param str name: column name to be found.

        :param type[pyrin.database.model.base.BaseEntity] scope: entities to search
                                                                 for the column.

        :keyword bool readable: specifies that any column or attribute
                                which has `allow_read=False` or its name
                                starts with underscore `_`, should not
                                be included in filtering.
                                defaults to True if not provided.

        :keyword list[InstrumentedAttribute | hybrid_property] exclude: list of columns
                                                                        to be excluded if
                                                                        have been found.

        :rtype: sqlalchemy.orm.attributes.InstrumentedAttribute |
                sqlalchemy.ext.hybrid.hybrid_property
        """

        exclude = options.get('exclude')
        exclude = misc_utils.make_iterable(exclude)
        readable = options.get('readable', True)
        for entity in scope:
            all_columns = None
            if readable is False:
                all_columns = entity.primary_key_columns + \
                              entity.foreign_key_columns + \
                              entity.all_columns + \
                              entity.expression_level_hybrid_properties
            else:
                all_columns = entity.readable_primary_key_columns + \
                              entity.readable_foreign_key_columns + \
                              entity.readable_columns + \
                              entity.expression_level_hybrid_properties

            if name in all_columns:
                column = entity.get_attribute(name)
                if column not in exclude:
                    return column

        return None

    def _add_expression(self, expressions, column,
                        name, to_be_removed, filters):
        """
        adds the relevant expression for given column into given expressions list.

        :param list expressions: list of expressions to add new expression to it.

        :param InstrumentedAttribute | hybrid_property column: column to add
                                                               expression for it.

        :param str name: relevant column name for filtering.
        :param list[str] to_be_removed: a list to add names to it if they used in expression.
        :param dict filters: filters to be applied.
        """

        is_attribute = isinstance(column, InstrumentedAttribute)
        collection_type = None
        python_type = None
        if is_attribute:
            collection_type, python_type = column.get_python_type()

        if name in filters:
            value = filters.get(name)
            to_be_removed.append(name)
            if not is_attribute and isinstance(value, str):
                expressions.append(func.lower(column).contains(value.lower()))

            elif python_type is str and not isinstance(value, LIST_TYPES) \
                    and collection_type is None:
                expressions.append(column.icontains(str(value)))

            else:
                sqlalchemy_utils.add_comparison_clause(expressions, column, value)

        if is_attribute and self.is_range_supported_column(column):
            from_name, to_name = self.get_range_filter_names(name)
            if from_name in filters or to_name in filters:
                from_value = filters.get(from_name)
                to_value = filters.get(to_name)
                to_be_removed.extend([from_name, to_name])
                if python_type is datetime:
                    sqlalchemy_utils.add_datetime_range_clause(expressions, column,
                                                               from_value, to_value)
                elif python_type is str:
                    sqlalchemy_utils.add_string_range_clause(expressions, column,
                                                             from_value, to_value)
                else:
                    sqlalchemy_utils.add_range_clause(expressions, column,
                                                      from_value, to_value)

    def filter(self, filters, *entity, **options):
        """
        gets filtering expressions for given entity types based on given filters.

        **NOTE:**

        if a filter name is available in both `labeled_filters` and columns of
        entities, the value of `labeled_filters` will be used.

        if a filter name is available in columns of more than one entity, the
        first found entity will be used.

        :param dict filters: filters to be applied.

        :param type[pyrin.database.model.base.BaseEntity] entity: entity class type to
                                                                  be used for filtering.

        :keyword list[InstrumentedAttribute | hybrid_property] ignore: columns to be ignored
                                                                       from filtering. this
                                                                       only has effect on
                                                                       `entity` and will be
                                                                       ignored for
                                                                       `labeled_filters`.

        :keyword bool remove: remove all keys that are applied as filter from filters input.
                              defaults to True if not provided.

        :keyword bool readable: specifies that any column or attribute
                                which has `allow_read=False` or its name
                                starts with underscore `_`, should not
                                be included in filtering.
                                defaults to True if not provided.
                                this only has effect on `entity` and will be
                                ignored for `labeled_filters`.

        :keyword dict labeled_filters: a dict containing all columns that should have
                                       a different filter name than their actual column
                                       name. for example:
                                       {'city_name': CityEntity.name,
                                        'person_name': PersonEntity.name}

        :keyword bool ignore_duplicates: specifies that if multiple labeled filters
                                         point to the same column and have the same value
                                         in provided filters, only add one of them in
                                         where clause. defaults to True if not provided.
                                         the main usage for this is in admin page which
                                         produces inclusive filters dynamically.

        :returns: list of expressions for filtering.
        :rtype: list
        """

        labeled_filters = options.get('labeled_filters') or {}
        ignore_duplicates = options.get('ignore_duplicates', True)
        remove = options.get('remove', True)
        ignore = options.get('ignore')
        ignore = misc_utils.make_iterable(ignore)
        ignore_names = [item.key for item in ignore]
        expressions = []
        to_be_removed = []
        filters_copy = dict(**filters)

        if len(labeled_filters) > 0:
            added = dict()
            for name in filters_copy:
                if name not in to_be_removed:
                    column_name = self._get_related_column_name(name)
                    column = labeled_filters.get(column_name)
                    if column is not None:
                        if ignore_duplicates is True and not self._is_range_filter_name(name):
                            added_values = added.setdefault(column, [])
                            value = filters_copy.get(name)
                            if value in added_values:
                                to_be_removed.append(name)
                                continue
                            else:
                                added_values.append(value)

                        self._add_expression(expressions, column, column_name,
                                             to_be_removed, filters_copy)

            for item in to_be_removed:
                filters_copy.pop(item, None)

        if len(entity) > 0:
            options.update(exclude=list(labeled_filters.values()))
            for name, value in filters_copy.items():
                if name not in to_be_removed:
                    column_name = self._get_related_column_name(name)
                    if column_name not in ignore_names:
                        column = self._get_related_column(column_name, *entity, **options)
                        if column is not None:
                            self._add_expression(expressions, column, column_name,
                                                 to_be_removed, filters_copy)

        if remove is not False:
            for item in to_be_removed:
                filters.pop(item, None)

        return expressions

    def get_range_filter_names(self, name):
        """
        gets range filter names for given name.

        it returns a tuple of two items. if name is `created_at`, then
        it returns (`from_created_at`, `to_created_at`) as range filters.

        :param str name: column name.

        :returns: tuple[str lower_bound_name, str upper_bound_name]
        :rtype: tuple[str, str]
        """

        from_range = '{prefix}{field}'.format(prefix=self.FROM_KEYWORD_PREFIX, field=name)
        to_range = '{prefix}{field}'.format(prefix=self.TO_KEYWORD_PREFIX, field=name)
        return from_range, to_range

    def is_range_supported_type(self, type_):
        """
        gets a value indicating that range operations are supported for given type.

        :param type type_: type to be checked for range support.

        :rtype: bool
        """

        return type_ in self.RANGE_SUPPORTED_TYPES

    def is_range_supported_column(self, column):
        """
        gets a value indicating that range operations are supported for given column.

        columns which their python type is not in `RANGE_SUPPORTED_TYPES`,
        do not support range operations.

        :param sqlalchemy.orm.attributes.InstrumentedAttribute column: column to be checked
                                                                       for range support.

        :rtype: bool
        """

        __, python_type = column.get_python_type()
        return self.is_range_supported_type(python_type)



# -*- coding: utf-8 -*-
"""
request wrappers enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class DeserializationTypeEnum(CoreEnum):
    """
    deserialization type enum.
    """

    FORM_DATA = 'form data'
    QUERY_STRINGS = 'query strings'
    URL_PARAMS = 'url params'


class CookieSameSiteEnum(CoreEnum):
    """
    cookie same-site enum.
    """

    STRICT = 'Strict'
    LAX = 'Lax'
    NONE = 'None'



# -*- coding: utf-8 -*-
"""
admin services module.
"""

from pyrin.application.services import get_component
from pyrin.admin import AdminPackage


def register(instance, **options):
    """
    registers the provided instance into available admin pages.

    :param pyrin.admin.interface.AbstractAdminPage instance: admin page instance.

    :keyword bool replace: specifies that if another admin page with the same name
                           or the same entity exists, replace it.
                           defaults to False if not provided and raises an error.

    :raises InvalidAdminPageTypeError: invalid admin page type error.
    :raises DuplicatedAdminPageError: duplicated admin page error.
    """

    return get_component(AdminPackage.COMPONENT_NAME).register(instance, **options)


def try_get_admin_page(entity):
    """
    gets the admin page for given entity class.

    it returns None if admin page does not exist.

    :param type[pyrin.database.model.base.BaseEntity] entity: the entity class of
                                                              admin page to be get.

    :rtype: pyrin.admin.interface.AbstractAdminPage
    """

    return get_component(AdminPackage.COMPONENT_NAME).try_get_admin_page(entity)


def is_admin_enabled():
    """
    gets a value indicating that admin api is enabled.

    :rtype: bool
    """

    return get_component(AdminPackage.COMPONENT_NAME).is_admin_enabled()


def has_admin(entity):
    """
    gets a value indicating that given entity class has admin page.

    :param type[pyrin.database.model.base.BaseEntity] entity: entity class.

    :rtype: bool
    """

    return get_component(AdminPackage.COMPONENT_NAME).has_admin(entity)


def get_admin_base_url():
    """
    gets admin base url.

    :rtype: str
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_admin_base_url()


def get_admin_configurations():
    """
    gets the admin api configurations.

    :returns: dict(bool enabled: enable admin api,
                   bool authenticated: admin api access type,
                   str url: admin api base url)
    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_admin_configurations()


def get_default_category():
    """
    gets the default category to be used for admin pages without category.

    :rtype: str
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_default_category()


def get(register_name, pk):
    """
    gets an entity with given primary key.

    :param str register_name: register name of admin page.
    :param object pk: primary key of entity to be get.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.
    :raises EntityNotFoundError: entity not found error.

    :rtype: pyrin.database.model.base.BaseEntity
    """

    return get_component(AdminPackage.COMPONENT_NAME).get(register_name, pk)


def find(register_name, **filters):
    """
    performs find on given admin page and returns the result.

    :param str register_name: register name of admin page.

    :keyword **filters: all filters to be passed to related admin page.

    :rtype: list[ROW_RESULT]
    """

    return get_component(AdminPackage.COMPONENT_NAME).find(register_name, **filters)


def create(register_name, **data):
    """
    performs create on given admin page.

    :param str register_name: register name of admin page.

    :keyword **data: all data to be passed to related admin page for data creation.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.

    :rtype: object
    """

    return get_component(AdminPackage.COMPONENT_NAME).create(register_name, **data)


def update(register_name, pk, **data):
    """
    performs update on given admin page.

    :param str register_name: register name of admin page.
    :param object pk: entity primary key to be updated.

    :keyword **data: all data to be passed to related admin page for data creation.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.
    :raises EntityNotFoundError: entity not found error.
    """

    return get_component(AdminPackage.COMPONENT_NAME).update(register_name, pk, **data)


def remove(register_name, pk):
    """
    performs remove on given admin page.

    :param str register_name: register name of admin page.
    :param object pk: entity primary key to be removed.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.
    """

    return get_component(AdminPackage.COMPONENT_NAME).remove(register_name, pk)


def remove_bulk(register_name, pk):
    """
    performs remove bulk on given admin page.

    :param str register_name: register name of admin page.
    :param object | list[object] pk: entity primary keys to be removed.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.
    """

    return get_component(AdminPackage.COMPONENT_NAME).remove_bulk(register_name, pk)


def remove_all(register_name):
    """
    performs remove all on given admin page.

    :param str register_name: register name of admin page.

    :raises AdminOperationNotAllowedError: admin operation not allowed error.
    """

    return get_component(AdminPackage.COMPONENT_NAME).remove_all(register_name)


def populate_main_metadata():
    """
    populates all admin pages main metadata.
    """

    return get_component(AdminPackage.COMPONENT_NAME).populate_main_metadata()


def get_main_metadata():
    """
    gets all admin pages main metadata.

    :raises AdminPagesHaveNotLoadedError: admin pages have not loaded error.

    :returns: dict(list pages,
                   dict configs)
    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_main_metadata()


def get_find_metadata(register_name):
    """
    gets the find metadata for given admin page.

    :param str register_name: register name of admin page.

    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_find_metadata(register_name)


def get_create_metadata(register_name):
    """
    gets the create metadata for given admin page.

    :param str register_name: register name of admin page.

    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_create_metadata(register_name)


def get_update_metadata(register_name):
    """
    gets the update metadata for given admin page.

    :param str register_name: register name of admin page.

    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_update_metadata(register_name)


def register_name_for(entity):
    """
    gets the admin page register name for given entity class.

    it raises an error if the given entity does not have an admin page.

    :param type[pyrin.database.model.base.BaseEntity] entity: the entity class of
                                                              admin page to get its
                                                              register name.

    :raises AdminPageNotFoundError: admin page not found error.

    :rtype: str
    """

    return get_component(AdminPackage.COMPONENT_NAME).register_name_for(entity)


def url_for(register_name):
    """
    gets the base url for given admin page.

    :param str register_name: admin page register name.

    :rtype: str
    """

    return get_component(AdminPackage.COMPONENT_NAME).url_for(register_name)


def get_list_field_type(form_field_type):
    """
    gets the equivalent list field type for given form field type.

    it may return None.

    :param str form_field_type: form field type to get its list field type.
    :enum form_field_type:
        BOOLEAN = 'boolean'
        DATE = 'date'
        DATETIME = 'datetime'
        TIME = 'time'
        EMAIL = 'email'
        FILE = 'file'
        NUMBER = 'number'
        PASSWORD = 'password'
        TELEPHONE = 'telephone'
        STRING = 'string'
        TEXT = 'text'
        URL = 'url'
        UUID = 'uuid'
        IPV4 = 'ipv4'
        IPV6 = 'ipv6'
        OBJECT = 'object'

    :rtype: str
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_list_field_type(form_field_type)


def populate_caches():
    """
    populates required caches of all registered admin pages.

    :returns: count of registered admin pages
    :rtype: int
    """

    return get_component(AdminPackage.COMPONENT_NAME).populate_caches()


def get_configs():
    """
    gets the required configs of admin api.

    :returns: dict(str panel_name,
                   str page_key,
                   str page_size_key,
                   str ordering_key,
                   str locale_key,
                   str timezone_key,
                   str search_param,
                   str hidden_pk_name)
    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).get_configs()


def login(username, password, **options):
    """
    logs in an internal user with given info into admin panel.

    :param str username: username.
    :param str password: password.

    :raises ProvidedUsernameOrPasswordAreIncorrect: provided username or
                                                    password are incorrect.

    :returns: dict(str access_token)
    :rtype: dict
    """

    return get_component(AdminPackage.COMPONENT_NAME).login(username, password, **options)



# -*- coding: utf-8 -*-
"""
audit security package.
"""

from pyrin.packaging.base import Package


class AuditSecurityPackage(Package):
    """
    audit security package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.security.authentication.handlers',
               'pyrin.security.authorization.handlers']



# -*- coding: utf-8 -*-
"""
authentication handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.security.exceptions import AuthenticationFailedError


class AuthenticationHandlersException(CoreException):
    """
    authentication handlers exception.
    """
    pass


class AuthenticationHandlersBusinessException(CoreBusinessException,
                                              AuthenticationHandlersException):
    """
    authentication handlers business exception.
    """
    pass


class AccessAndRefreshTokensDoesNotBelongToSameUserError(AuthenticationFailedError,
                                                         AuthenticationHandlersBusinessException):
    """
    access and refresh tokens does not belong to same user error.
    """
    pass


class InvalidTokenAuthenticatorError(AuthenticationFailedError,
                                     AuthenticationHandlersBusinessException):
    """
    invalid token authenticator error.
    """
    pass


class AuthenticatorNameIsRequiredError(AuthenticationHandlersException):
    """
    authenticator name is required error.
    """
    pass


class AccessTokenRequiredError(AuthenticationFailedError,
                               AuthenticationHandlersBusinessException):
    """
    access token required error.
    """
    pass


class RefreshTokenRequiredError(AuthenticationFailedError,
                                AuthenticationHandlersBusinessException):
    """
    refresh token required error.
    """
    pass


class InvalidAccessTokenError(AuthenticationFailedError,
                              AuthenticationHandlersBusinessException):
    """
    invalid access token error.
    """
    pass


class InvalidRefreshTokenError(AuthenticationFailedError,
                               AuthenticationHandlersBusinessException):
    """
    invalid refresh token error.
    """
    pass


class UserCredentialsRevokedError(AuthenticationFailedError,
                                  AuthenticationHandlersBusinessException):
    """
    user credentials revoked error.
    """
    pass


class InvalidUserError(AuthenticationHandlersException):
    """
    invalid user error.
    """
    pass


class InvalidUserIdentityError(AuthenticationHandlersException):
    """
    invalid user identity error.
    """
    pass


class ProvidedUsernameOrPasswordAreIncorrect(AuthenticationFailedError,
                                             AuthenticationHandlersBusinessException):
    """
    provided username or password are incorrect.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching admin module.
"""

from pyrin.admin.decorators import admin
from pyrin.admin.page.base import AdminPage
from pyrin.caching.models import CacheItemEntity


@admin()
class CacheItemAdmin(AdminPage):
    entity = CacheItemEntity
    register_name = 'cache-items'
    name = 'Cache Item'
    create_permission = False
    update_permission = False
    list_fields = (CacheItemEntity.id, CacheItemEntity.cache_name,
                   CacheItemEntity.shard_name, CacheItemEntity.version,
                   CacheItemEntity.key, CacheItemEntity.created_at)



# -*- coding: utf-8 -*-
"""
admin security package.
"""

from pyrin.packaging.base import Package


class AdminSecurityPackage(Package):
    """
    admin security package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.security.authentication.handlers',
               'pyrin.security.authorization.handlers']



# -*- coding: utf-8 -*-
"""
admin page base module.
"""

import inspect

from sqlalchemy.sql.elements import Label, and_, or_
from sqlalchemy.orm import InstrumentedAttribute

import pyrin.utils.path as path_utils
import pyrin.utils.string as string_utils
import pyrin.utils.sqlalchemy as sqla_utils
import pyrin.utils.misc as misc_utils
import pyrin.utils.dictionary as dict_utils
import pyrin.admin.services as admin_services
import pyrin.filtering.services as filtering_services
import pyrin.validator.services as validator_services
import pyrin.security.session.services as session_services
import pyrin.database.model.services as model_services
import pyrin.configuration.services as config_services
import pyrin.database.paging.services as paging_services
import pyrin.utilities.string.normalizer.services as string_normalizer_services

from pyrin.core.globals import _
from pyrin.core.structs import SecureList
from pyrin.admin.interface import AbstractAdminPage
from pyrin.admin.page.schema import AdminSchema
from pyrin.core.globals import SECURE_TRUE
from pyrin.admin.page.mixin import AdminPageCacheMixin
from pyrin.caching.mixin.decorators import fast_cache
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.paging.paginator import SimplePaginator
from pyrin.database.services import get_current_store
from pyrin.database.model.base import BaseEntity
from pyrin.logging.contexts import suppress
from pyrin.security.session.enumerations import RequestContextEnum
from pyrin.validator.exceptions import ValidationError
from pyrin.admin.page.enumerations import TableTypeEnum, PaginationTypeEnum, \
    PaginationPositionEnum, FormatEnum, MonthFormatEnum, ButtonTypeEnum, \
    LinkTypeEnum, HourCycleEnum
from pyrin.admin.page.exceptions import InvalidListFieldError, ListFieldRequiredError, \
    InvalidMethodNameError, InvalidAdminEntityTypeError, AdminNameRequiredError, \
    AdminRegisterNameRequiredError, RequiredValuesNotProvidedError, \
    CompositePrimaryKeysNotSupportedError, DuplicateListFieldNamesError, \
    InvalidListSearchFieldError, DuplicateListSearchFieldNamesError, EntityNotFoundError, \
    InvalidListFieldNameError, ExtraDataFieldsAndEntityFieldsOverlapError


class AdminPage(AbstractAdminPage, AdminPageCacheMixin):
    """
    admin page class.

    all admin pages must be subclassed from this.
    """

    # ===================== REQUIRED CONFIGS ===================== #

    # the entity class that this admin page represents.
    entity = None

    # name of this admin page to be used for registration.
    # the register name is case-insensitive and must be unique for each admin page.
    # this name will also be used for generating admin page api routes.
    register_name = None

    # singular name of this admin page for representation.
    name = None

    # ===================== LIST CONFIGS ===================== #

    # columns to show in list view. it could be a column attribute, an expression
    # level hybrid property or a string method name of this admin page.
    # the method should accept a single positional argument as current row.
    # for example: (UserEntity.id, UserEntity.fullname, 'title', UserDetailEntity.age)
    list_fields = ()

    # columns that will be selected from database and will be used to compute
    # another field's value but they will be removed from the final result and
    # will not be returned to client.
    # it could be a column attribute or an expression level hybrid property.
    # for example: (UserEntity.id, UserEntity.fullname, UserDetailEntity.age)
    list_temp_fields = ()

    # specifies that if 'list_fields' is not provided only show readable
    # columns of the entity in list view.
    list_only_readable = True

    # specifies that if 'list_fields' are not provided also show pk columns in list view.
    # note that if the admin page has get permission, pk columns will always be added.
    list_pk = True

    # specifies that if 'list_fields' are not provided also show fk columns in list view.
    list_fk = True

    # specifies that if 'list_fields' are not provided also show expression
    # level hybrid property columns in list view.
    list_expression_level_hybrid_properties = True

    # list of default ordering columns.
    # it must be string names from 'list_fields' or 'list_temp_fields'.
    # for example: ('first_name', '-last_name')
    list_ordering = ()

    # specifies that each row must have an index in it.
    list_indexed = False

    # list index name to be added to each row.
    list_index_name = None

    # start index of rows.
    list_start_index = 1

    # columns to show in list filter.
    list_filters = ()

    # paginate results in list view.
    list_paged = True

    # records per page.
    # it could not be more than 'list_max_page_size', otherwise it
    # will be corrected silently.
    list_page_size = None

    # max records per page.
    # it could not be more than 'max_page_size' from 'database'
    # config store, otherwise it will be corrected silently.
    list_max_page_size = None

    # pagination type to be used on client if paging is enabled for this admin page.
    # it should be from 'PaginationTypeEnum' values.
    list_pagination_type = PaginationTypeEnum.NORMAL

    # pagination position to be used on client if paging is enabled for this admin page.
    # it should be from 'PaginationPositionEnum' values.
    list_pagination_position = PaginationPositionEnum.BOTTOM

    # a value to be shown when the relevant data is null.
    list_null_value = '-'

    # a dict containing field names and their list field type for fields which their
    # type could not be detected automatically. for example hybrid property fields
    # or fields which their value is coming from a method of this admin page.
    # note that the provided types must be from 'ListFieldTypeEnum' values.
    # for example: {'is_viewed': ListFieldTypeEnum.BOOLEAN}
    list_extra_field_types = {}

    # a value to define client-side table type.
    # it should be from 'TableTypeEnum' values.
    list_table_type = TableTypeEnum.DENSE

    # enable exporting the currently active page into pdf or csv file.
    list_enable_export = True

    # a file name to be used for exporting data. for example: Users
    # if not provided defaults to this admin page's plural name.
    list_export_name = None

    # create a link to related admin detail page for all pk columns.
    # this only has effect if the related admin page has get permission.
    list_link_pk = True

    # create a link to related admin detail page for all fk columns.
    # this only has effect if the related admin page has get permission.
    list_link_fk = True

    # let user select which columns to show on list page.
    list_column_selection = True

    # let user re-order table columns on list page.
    list_column_ordering = True

    # let user group results by columns on list page.
    list_grouping = False

    # specifies that client should render a search box to let user filter the
    # results by any of columns which are involved in this admin page list view.
    list_search = True

    # a number in milliseconds to be used as debounce interval in client search box.
    # note that if you set it to lower numbers, it will result in a service call
    # for each keystroke in search box.
    list_search_debounce_interval = 700

    # extra columns to let user filter by their value in list view.
    # it must be a column attribute or a labeled column.
    # this is useful if you want to add some extra fields for searching but do not
    # want to add these columns in 'list_fields'.
    # the main usage for this attribute is to set labeled columns for columns
    # of different entities which have the same name.
    # for example:
    # (UserEntity.last_name, UserEntity.name, CityEntity.name.label('city_name'))
    list_search_fields = ()

    # format to render datetime fields on list page.
    list_datetime_format = dict(year=FormatEnum.NUMERIC,
                                month=MonthFormatEnum.SHORT,
                                day=FormatEnum.NUMERIC,
                                hour=FormatEnum.TWO_DIGIT,
                                minute=FormatEnum.TWO_DIGIT,
                                second=FormatEnum.TWO_DIGIT,
                                hourCycle=HourCycleEnum.H23)

    # format to render date fields on list page.
    list_date_format = dict(year=FormatEnum.NUMERIC,
                            month=MonthFormatEnum.SHORT,
                            day=FormatEnum.NUMERIC)

    # format to render time fields on list page.
    list_time_format = dict(hour=FormatEnum.TWO_DIGIT,
                            minute=FormatEnum.TWO_DIGIT,
                            second=FormatEnum.TWO_DIGIT,
                            hourCycle=HourCycleEnum.H23)

    # a locale to be used to render date and time fields on list page.
    # for example: 'en-US', 'fa' or ...
    # if not provided, the locale on client browser will be used.
    list_locale = None

    # a value to be used as max body height for table in list view.
    # for example: '800px'
    # if not set, the client will automatically choose the best height
    # based on the screen resolution.
    list_max_body_height = None

    # ===================== SERVICE CONFIGS ===================== #

    # a service to be used for create operation.
    # if not set, the default create operation of this admin page will be used.
    # the create service must also accept keyword arguments.
    create_service = None

    # a service to be used for update operation.
    # if not set, the default update operation of this admin page will be used.
    # the update service must accept a positional argument at the beginning as
    # the primary key of the related entity and also accept keyword arguments.
    update_service = None

    # a service to be used for remove operation.
    # if not set, the default remove operation of this admin page will be used.
    # the remove service must accept a single positional argument as the
    # primary key of the related entity.
    remove_service = None

    # ===================== PERMISSION CONFIGS ===================== #

    # specifies that this admin page has get permission.
    # note that if entity has composite primary key, it does not
    # have get permission and this value will be ignored.
    get_permission = True

    # specifies that this admin page has create permission.
    create_permission = True

    # specifies that this admin page has update permission.
    # note that if entity has composite primary key, it does not
    # have update permission and this value will be ignored.
    update_permission = True

    # specifies that this admin page has single or bulk remove permission.
    # note that if entity has composite primary key, it does not have single
    # or bulk remove permission and this value will be ignored.
    remove_permission = True

    # specifies that this admin page has remove all permission.
    remove_all_permission = True

    # ===================== OTHER CONFIGS ===================== #

    # the category name to register this admin page in it.
    # all admin pages with the same category will be grouped together.
    # the category name is case-insensitive.
    # it will be set to the package name of current admin page if not provided.
    category = None

    # plural name of this admin page for representation.
    # it will be automatically generated from 'name' if not provided.
    plural_name = None

    # extra field names that are required to be provided for create and are
    # optional for update but they are not a field of the entity itself.
    # in the form of:
    # {str field_name: type python_type}
    # for example:
    # {'password': str, 'join_date': datetime, 'age': int}
    # if the provided names have related validators, the
    # info of that validator will be sent to client.
    extra_data_fields = {}

    # ===================== INTERNAL CONFIGS ===================== #

    # paginator class to be used.
    paginator_class = SimplePaginator

    # the fully qualified name of find api.
    FIND_ENDPOINT = 'pyrin.admin.api.find'

    def __init__(self, *args, **options):
        """
        initializes an instance of AdminPage.

        :raises InvalidAdminEntityTypeError: invalid admin entity type error.
        :raises AdminRegisterNameRequiredError: admin register name required error.
        :raises AdminNameRequiredError: admin name required error.
        """

        super().__init__()

        if not inspect.isclass(self.entity) or not issubclass(self.entity, BaseEntity):
            raise InvalidAdminEntityTypeError('The entity for [{admin}] class '
                                              'must be a subclass of [{base}].'
                                              .format(admin=self, base=BaseEntity))

        if self.register_name in (None, '') or self.register_name.isspace():
            raise AdminRegisterNameRequiredError('The register name for '
                                                 '[{admin}] class is required.'
                                                 .format(admin=self))

        if self.name in (None, '') or self.name.isspace():
            raise AdminNameRequiredError('The name for [{admin}] class is required.'
                                         .format(admin=self))

        # list of method names of this admin page to be used for processing the list results.
        self._method_names = self._extract_method_names()
        self._schema = AdminSchema(self,
                                   indexed=self.list_indexed,
                                   start_index=self.list_start_index,
                                   index_name=self.list_index_name,
                                   exclude=self._get_list_temp_field_names())
        self._paginator = None
        if self.list_paged is True:
            self._paginator = self.paginator_class(self.FIND_ENDPOINT,
                                                   page_size=self._get_page_size(),
                                                   max_page_size=self._get_max_page_size())

    def _get_hidden_pk_name(self):
        """
        gets hidden pk name from configs.

        :rtype: str
        """

        return config_services.get_active('admin', 'hidden_pk_name')

    def _get_search_param(self):
        """
        gets search param name from configs.

        :rtype: str
        """

        return config_services.get_active('admin', 'search_param')

    @fast_cache
    def _get_list_export_name(self):
        """
        gets a file name to be used for exporting data on list page.

        :rtype: str
        """

        if self.list_export_name:
            return self.list_export_name

        return self.get_plural_name()

    def _get_column_title(self, name):
        """
        gets the column title of given field name for list page.

        :param str name: field name.

        :raises InvalidListFieldNameError: invalid list field name error.

        :rtype: str
        """

        if name is None:
            raise InvalidListFieldNameError('There are some fields in "list_fields" of '
                                            '[{admin}] class which does not have a name. '
                                            'use label to set a name for them.'
                                            .format(admin=self))

        name = name.replace('_', ' ')
        return name.upper()

    def _get_field_title(self, name):
        """
        gets the field title of given field name for detail and create page.

        :param str name: field name.

        :rtype: str
        """

        name = name.replace('_', ' ')
        return string_normalizer_services.title_case(name)

    @classmethod
    @fast_cache
    def _get_primary_key_name(cls):
        """
        gets the name of the primary key of this admin page's related entity.

        note that if the entity has a composite primary key, this method raises an error.

        :rtype: str
        """

        if len(cls.entity.primary_key_columns) == 1:
            return cls.entity.primary_key_columns[0]

        raise CompositePrimaryKeysNotSupportedError('Composite primary keys are not '
                                                    'supported for admin page.')

    @classmethod
    def _get_primary_key_holder(cls, pk):
        """
        gets a dict with the primary key name of this page's entity set to the given value.

        :param object pk: value to be set to primary key.

        :rtype: dict
        """

        pk_name = cls._get_primary_key_name()
        pk_holder = dict()
        pk_holder[pk_name] = pk

        return pk_holder

    def _is_valid_field(self, field):
        """
        gets a value indicating that the provided field is a valid field for list fields.

        :param object field: field to be checked.

        :rtype: bool
        """

        return hasattr(field, 'key')

    def _is_valid_method(self, name):
        """
        gets a value indicating that the provided name is a valid method name of this admin page.

        :param str name: method name.

        :rtype: bool
        """

        method = getattr(self, name, None)
        return callable(method)

    def _is_list_pk_required(self):
        """
        gets a value indicating that primary key column for list view is required.

        it returns True if this admin page has any of `get` or `remove` permissions.

        :rtype: bool
        """

        return self.has_get_permission() or self.has_remove_permission()

    @fast_cache
    def _extract_method_names(self):
        """
        extracts all valid method names from list fields.

        :rtype: tuple[str]
        """

        names = []
        if self.list_fields:
            for item in self.list_fields:
                if isinstance(item, str) and self._is_valid_method(item):
                    names.append(item)

        return tuple(names)

    def _extract_field_names(self, fields, allow_string=True):
        """
        extracts fields names form given fields

        :param list | tuple fields: list of fields to extract their names.

        :param bool allow_string: specifies that string fields should also
                                  be accepted. defaults to True if not provided.

        :raises InvalidListFieldError: invalid list field error.

        :rtype: tuple[str]
        """

        names = []
        for item in fields:
            if self._is_valid_field(item):
                names.append(item.key)
            elif allow_string is True and isinstance(item, str) \
                    and self._is_valid_method(item):
                names.append(item)
            else:
                message = 'Provided field [{field}] is not a valid value. ' \
                          'it must be a column attribute{sign} ' \
                          'expression level hybrid property{end}'

                if allow_string is True:
                    message = message.format(
                        field=str(item), sign=',',
                        end=' or a string representing a method name of [{admin}] class.')
                else:
                    message = message.format(field=str(item), sign=' or',
                                             end='.')

                raise InvalidListFieldError(message.format(admin=self))

        return tuple(names)

    @fast_cache
    def _get_list_field_names(self):
        """
        gets all list field names of this admin page.

        :raises InvalidListFieldError: invalid list field error.
        :raises DuplicateListFieldNamesError: duplicate list field names error.

        :rtype: tuple[str]
        """

        all_fields = ()
        if not self.list_fields:
            all_fields = self._get_default_list_fields()
        else:
            all_fields = list(self.list_fields)
            if self._is_list_pk_required():
                self._inject_primary_key(all_fields)

        all_fields = self._extract_field_names(all_fields, allow_string=True)
        if self.list_indexed is True:
            index_name = self.list_index_name
            if index_name in (None, ''):
                index_name = config_services.get('api', 'schema', 'index_name')

            all_fields = list(all_fields)
            all_fields.insert(0, index_name)
            all_fields = tuple(all_fields)

        duplicates = misc_utils.get_duplicates(all_fields)
        if duplicates:
            raise DuplicateListFieldNamesError('There are some duplicate field names '
                                               'in "list_fields" of [{admin}] class. '
                                               'this will make the find result incorrect. '
                                               'please remove duplicate fields or set '
                                               'unique labels for them: {duplicates}'
                                               .format(admin=self, duplicates=duplicates))

        return all_fields

    def _get_attribute(self, item):
        """
        gets the related instrumented or hybrid property attribute to given item.

        it may return the same value if it is already an instrumented or
        hybrid property attribute. it may return None if no related instrumented
        or hybrid property attribute could be found.

        :param InstrumentedAttribute | Label | hybrid_property item: item to get its
                                                                     related attribute.

        :rtype: sqlalchemy.orm.InstrumentedAttribute | sqlalchemy.ext.hybrid.hybrid_property
        """

        if isinstance(item, InstrumentedAttribute) or \
                sqla_utils.is_expression_level_hybrid_property(item):
            return item
        elif isinstance(item, Label):
            hybrid_property = getattr(item, '_hybrid_property', None)
            if hybrid_property is not None:
                return hybrid_property
            elif item.base_columns:
                columns = list(item.base_columns)
                if isinstance(columns[0], CoreColumn):
                    return model_services.get_instrumented_attribute(columns[0])

        return None

    def _get_list_field_column(self, name):
        """
        gets the relevant column or hybrid property attribute for given field name.

        it may return None if no related column or hybrid property could be found.

        :param str name: field name.

        :rtype: sqlalchemy.orm.InstrumentedAttribute | sqlalchemy.ext.hybrid.hybrid_property
        """

        selectable_fields = self._get_list_fields()
        for item in selectable_fields:
            if self._is_valid_field(item) and item.key == name:
                return self._get_attribute(item)

        return None

    @fast_cache
    def _get_link_methods(self):
        """
        gets all method names which should render a link on client list view.

        :rtype: tuple[str]
        """

        links = []
        for name in self.method_names:
            method = getattr(self, name)
            is_link = getattr(method, 'is_link', False)
            if is_link is True:
                links.append(name)

        return tuple(links)

    @fast_cache
    def _get_list_fields_to_column_map(self):
        """
        gets a dict of all list field names and their related columns.

        the column value may be None if the name does not belong to a
        column or hybrid property. for example method names.

        :returns: dict(InstrumentAttribute | hybrid_property name)
        :rtype: dict
        """

        result = dict()
        all_fields = self._get_list_field_names()
        for name in all_fields:
            attribute = self._get_list_field_column(name)
            result[name] = attribute

        return result

    @fast_cache
    def _get_list_search_fields_to_column_map(self):
        """
        gets a dict of all list search field names and their related columns.

        :raises InvalidListSearchFieldError: invalid list search field error.
        :raises DuplicateListSearchFieldNamesError: duplicate list search field names error.

        :returns: dict(InstrumentAttribute | hybrid_property name)
        :rtype: dict
        """

        if self.list_search_fields:
            names = []
            for item in self.list_search_fields:
                if not isinstance(item, (InstrumentedAttribute, Label)) and \
                        not sqla_utils.is_expression_level_hybrid_property(item):
                    raise InvalidListSearchFieldError('Provided field [{field}] is not a valid '
                                                      'list search field for [{admin}] class. '
                                                      'search fields must be a column attribute '
                                                      'or an expression level hybrid property.'
                                                      .format(admin=self, field=str(item)))

                names.append(item.key)

            duplicates = misc_utils.get_duplicates(names)
            if duplicates:
                raise DuplicateListSearchFieldNamesError('There are some duplicate field names '
                                                         'in "list_search_fields" of [{admin}] '
                                                         'class. please remove duplicate fields '
                                                         'or set unique labels for them: '
                                                         '{duplicates}'
                                                         .format(admin=self,
                                                                 duplicates=duplicates))

        result = dict()
        list_fields = self._get_list_fields_to_column_map()
        for name, column in list_fields.items():
            if column is not None and name != self._get_hidden_pk_name():
                result[name] = column

        if self.list_search_fields:
            for item in self.list_search_fields:
                attribute = self._get_attribute(item)
                if attribute is not None:
                    result[item.key] = attribute
                elif sqla_utils.is_expression_level_hybrid_property(item):
                    result[item.key] = item

        return result

    def _get_pk_info(self, column):
        """
        gets a dict containing pk info for given column if required.

        :param InstrumentedAttribute column: column to be checked.

        :returns: dict(bool is_pk,
                       str pk_register_name,
                       str pk_name)
        :rtype: dict
        """

        result = dict(is_pk=False, pk_register_name=None)
        if sqla_utils.is_expression_level_hybrid_property(column) or \
                not column.property or not column.property.columns:
            return result

        base_column = column.property.columns[0]
        is_pk = base_column.primary_key
        if is_pk is True:
            admin_page = admin_services.try_get_admin_page(column.class_)
            if admin_page is not None and admin_page.has_get_permission() is True:
                result.update(is_pk=is_pk,
                              pk_register_name=admin_page.get_register_name(),
                              pk_name=admin_page.get_singular_name())

        return result

    def _get_fk_info(self, column):
        """
        gets a dict containing fk info for given column if required.

        :param InstrumentedAttribute column: column to be checked.

        :returns: dict(bool is_fk,
                       str fk_register_name,
                       str fk_name)
        :rtype: dict
        """

        result = dict(is_fk=False, fk_register_name=None)
        if sqla_utils.is_expression_level_hybrid_property(column) or \
                not column.property or not column.property.columns:
            return result

        base_column = column.property.columns[0]
        is_fk = base_column.is_foreign_key
        if is_fk is True:
            foreign_keys = list(base_column.foreign_keys)
            if foreign_keys and foreign_keys[0].constraint:
                entity = sqla_utils.get_class_by_table(model_services.get_declarative_base(),
                                                       foreign_keys[0].constraint.referred_table,
                                                       raise_multi=False)

                if entity is not None:
                    admin_page = admin_services.try_get_admin_page(entity)
                    if admin_page is not None and admin_page.has_get_permission() is True:
                        result.update(is_fk=is_fk,
                                      fk_register_name=admin_page.get_register_name(),
                                      fk_name=admin_page.get_singular_name())

        return result

    def _get_extra_list_field_info(self, name):
        """
        gets a dict of extra info for given list field.

        it may return an empty dict.

        :param str name: list field name.

        :rtype: dict
        """

        result = dict()
        column_map = self._get_list_fields_to_column_map()
        attribute = column_map.get(name)
        type_ = None
        if attribute is not None:
            validator = validator_services.try_get_validator(attribute.class_, attribute)
            if validator is not None:
                info = validator.get_info()
                type_ = admin_services.get_list_field_type(info.get('form_field_type'))
                lookup = info.get('in_enum_lookup')
                if lookup:
                    result.update(lookup=lookup)

            result.update(self._get_pk_info(attribute))
            result.update(self._get_fk_info(attribute))

        elif self.list_extra_field_types:
            type_ = self.list_extra_field_types.get(name)

        if type_ is not None:
            result.update(type=type_)

        return result

    @fast_cache
    def _get_list_datasource_info(self):
        """
        gets datasource info to be used for list page.

        :rtype: tuple[dict]
        """

        results = []
        all_fields = self._get_list_field_names()
        sortable_fields = self._get_sortable_fields()
        link_methods = self._get_link_methods()
        hidden_pk_name = self._get_hidden_pk_name()
        for item in all_fields:
            info = dict(field=item,
                        title=self._get_column_title(item),
                        sorting=item in sortable_fields,
                        emptyValue=self.list_null_value,
                        is_link=item in link_methods)

            if item in link_methods or item == hidden_pk_name:
                info.update(export=False)

            if item == hidden_pk_name:
                info.update(hidden=True, hiddenByColumnsButton=True)
            else:
                extra_info = self._get_extra_list_field_info(item)
                info.update(extra_info)

            results.append(info)

        return tuple(results)

    @fast_cache
    def _get_primary_keys(self):
        """
        gets all primary key attributes of this admin page's related entity.

        :rtype: tuple[InstrumentedAttribute]
        """

        return tuple(self.entity.get_attribute(name)
                     for name in self.entity.primary_key_columns)

    def _inject_primary_key(self, fields):
        """
        injects primary key attribute of related entity into given list.

        this is required if the current admin page has any
        of get, remove or update permissions. note that only single
        primary keys are allowed. otherwise it raises an error.

        :param list fields: fields to inject primary key into it.

        :raises CompositePrimaryKeysNotSupportedError: composite primary keys not supported error.
        """

        if not self._has_single_primary_key():
            raise CompositePrimaryKeysNotSupportedError('Composite primary keys are not '
                                                        'supported for admin page.')

        primary_key = self._get_primary_keys()[0]
        fields.append(primary_key.label(self._get_hidden_pk_name()))

    @fast_cache
    def _get_list_temp_field_names(self):
        """
        gets all list temp field names of this admin page.

        :raises InvalidListFieldError: invalid list field error.

        :rtype: tuple[str]
        """

        if not self.list_temp_fields:
            return tuple()

        return self._extract_field_names(self.list_temp_fields, allow_string=False)

    @fast_cache
    def _get_default_list_fields(self):
        """
        gets defaults list fields for this admin page.

        :rtype: tuple
        """

        primary_keys = []
        primary_key_names = ()
        foreign_keys = []
        foreign_key_names = ()
        columns = []
        column_names = ()
        expression_level_hybrid_properties = []
        expression_level_hybrid_property_names = ()

        if self.list_only_readable is True:
            column_names = self.entity.readable_columns
            if self.list_pk is True:
                primary_key_names = self.entity.readable_primary_key_columns

            if self.list_fk is True:
                foreign_key_names = self.entity.readable_foreign_key_columns
        else:
            column_names = self.entity.all_columns
            if self.list_pk is True:
                primary_key_names = self.entity.primary_key_columns

            if self.list_fk is True:
                foreign_key_names = self.entity.foreign_key_columns

        if self.list_expression_level_hybrid_properties is True:
            expression_level_hybrid_property_names = \
                self.entity.expression_level_hybrid_properties

        for pk in primary_key_names:
            primary_keys.append(self.entity.get_attribute(pk))

        for fk in foreign_key_names:
            foreign_keys.append(self.entity.get_attribute(fk))

        for column in column_names:
            columns.append(self.entity.get_attribute(column))

        for hybrid_property in expression_level_hybrid_property_names:
            attribute = self.entity.get_attribute(hybrid_property)
            expression_level_hybrid_properties.append(attribute.label(attribute.key))

        primary_keys.extend(foreign_keys)
        primary_keys.extend(columns)
        primary_keys.extend(expression_level_hybrid_properties)

        if self._is_list_pk_required():
            self._inject_primary_key(primary_keys)

        return tuple(primary_keys)

    @fast_cache
    def _get_list_fields(self):
        """
        gets all fields from `list_fields` that are column or hybrid property.

        :rtype: tuple
        """

        if not self.list_fields:
            return self._get_default_list_fields()

        results = []
        for item in self.list_fields:
            if self._is_valid_field(item):
                if sqla_utils.is_expression_level_hybrid_property(item):
                    labeled_item = item.label(item.key)
                    setattr(labeled_item, '_hybrid_property', item)
                    results.append(labeled_item)
                else:
                    results.append(item)

        if self._is_list_pk_required():
            self._inject_primary_key(results)

        return tuple(results)

    @fast_cache
    def _get_list_temp_fields(self):
        """
        gets all fields from `list_temp_fields`.

        :rtype: tuple
        """

        if not self.list_temp_fields:
            return tuple()

        results = [item for item in self.list_temp_fields if self._is_valid_field(item)]
        return tuple(results)

    @fast_cache
    def _get_selectable_fields(self):
        """
        gets all selectable fields of this admin page.

        :raises ListFieldRequiredError: list field required error.

        :rtype: tuple
        """

        results = self._get_list_fields() + self._get_list_temp_fields()
        if len(results) <= 0:
            raise ListFieldRequiredError('At least a single column attribute or '
                                         'expression level hybrid property must be '
                                         'provided in list fields or list temp fields '
                                         'of [{admin}] class.'.format(admin=self))

        return results

    def _perform_joins(self, query, **options):
        """
        performs joins on given query and returns a new query object.

        this method is intended to be overridden in subclasses if you want to
        provide columns of other entities in `list_fields` or `list_temp_fields`.

        :param CoreQuery query: query instance.

        :rtype: CoreQuery
        """

        return query

    def _prepare_inclusive_filters(self, search_text, filters, labeled_filters):
        """
        prepares inclusive filters.

        :param str search_text: search text.
        :param dict filters: filters which have been provided by client.
        :param dict labeled_filters: dict of filter names to column map.
        """

        for name, column in labeled_filters.items():
            value = None
            if sqla_utils.is_expression_level_hybrid_property(column):
                value = search_text
            else:
                with suppress(ValidationError, log=False):
                    value = validator_services.validate_field(column.class_, column,
                                                              search_text, for_find=True)

            if value is not None:
                filters[name] = value

    def _filter_query(self, query, **filters):
        """
        filters given query and returns a new query object.

        :param CoreQuery query: query instance.

        :rtype: CoreQuery
        """

        labeled_filters = self._get_list_search_fields_to_column_map()
        search_text = filters.pop(self._get_search_param(), None)
        type_ = and_
        if self.list_search is True and search_text not in (None, ''):
            type_ = or_
            self._prepare_inclusive_filters(search_text, filters, labeled_filters)

        expressions = filtering_services.filter(filters, labeled_filters=labeled_filters)
        return query.filter(type_(*expressions))

    def _validate_filters(self, filters):
        """
        validates given filters.

        :param dict filters: filters to be validated.

        :raises ValidationError: validation error.
        """

        validator_services.validate_for_find(self.entity, filters)

    def _perform_order_by(self, query, **filters):
        """
        performs order by on given query and returns a new query object.

        :param CoreQuery query: query instance.

        :keyword str | list[str] order_by: order by columns.
                                           if not provided, defaults to `ordering`
                                           fields of this admin class.
        :rtype: CoreQuery
        """

        filters.update(labeled_columns=SecureList(self._get_list_labels()))
        force_order = list(self.list_ordering or [])
        force_order.extend(self.entity.primary_key_columns)
        return query.safe_order_by(self._get_list_entities(), *force_order, **filters)

    @fast_cache
    def _get_page_size(self):
        """
        gets page size of this admin page.

        :rtype: int
        """

        max_page_size = self._get_max_page_size()
        default_page_size = config_services.get('database', 'paging', 'default_page_size')
        page_size = self.list_page_size
        if page_size is None or page_size < 1 or page_size > max_page_size:
            page_size = min(default_page_size, max_page_size)

        return page_size

    @fast_cache
    def _get_max_page_size(self):
        """
        gets max page size of this admin page.

        :rtype: int
        """

        global_max_page_size = config_services.get('database', 'paging', 'max_page_size')
        max_page_size = self.list_max_page_size
        if max_page_size is None or max_page_size < 1 or max_page_size > global_max_page_size:
            max_page_size = global_max_page_size

        return max_page_size

    def _get_paginator(self):
        """
        gets paginator for this admin page.

        :rtype: PaginatorBase
        """

        return self._paginator.copy()

    def _inject_paginator(self, filters):
        """
        injects this admin page's paginator into current request context.

        :param dict filters: view function inputs.
        """

        paginator = self._get_paginator()
        paging_services.inject_paginator(paginator, filters)

    def _paginate_query(self, query, **filters):
        """
        paginates given query and returns a new query object.

        it may return the same query if pagination is disabled for this admin page.

        :param CoreQuery query: query instance.

        :keyword CoreColumn column: column to be used in count function.
                                    defaults to `*` if not provided.
                                    this is only used if `inject_total` is
                                    provided and a single query could be
                                    produced for count.

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.
                                note that `distinct` will only be
                                used if `column` is also provided.

        :keyword int __limit__: limit value.
        :keyword int __offset__: offset value.

        :rtype: CoreQuery
        """

        if self.list_paged is True:
            self._inject_paginator(filters)
            return query.paginate(inject_total=SECURE_TRUE, **filters)

        return query

    @fast_cache
    def _get_page_size_options(self):
        """
        gets a tuple of page size options of this admin page.

        :rtype: tuple[int] | tuple[int, int]
        """

        page_size = self._get_page_size()
        max_page_size = self._get_max_page_size()
        if page_size == max_page_size:
            return page_size,

        return page_size, max_page_size

    def _process_find_results(self, results, **options):
        """
        processes the given results and returns a list of serialized values.

        :param list[ROW_RESULT] results: results to be processed.

        :rtype: list[dict]
        """

        paginator = session_services.get_request_context(RequestContextEnum.PAGINATOR)
        return self._schema.filter(results, paginator=paginator)

    @classmethod
    def _has_single_primary_key(cls):
        """
        gets a value indicating that the related entity has a single primary key.

        :rtype: bool
        """

        return len(cls.entity.primary_key_columns) == 1

    @classmethod
    def _validate_extra_fields(cls, data):
        """
        validates that all extra required fields are available in data.

        :param dict data: data to be validated.

        :raises RequiredValuesNotProvidedError: required values not provided error.
        """

        if not cls.extra_data_fields:
            return

        not_present = []
        for name in cls.extra_data_fields:
            if data.get(name) is None:
                not_present.append(name)

        not_present = set(not_present)
        if len(not_present) > 0:
            raise RequiredValuesNotProvidedError(_('These values are required: {values}')
                                                 .format(values=list(not_present)))

    @classmethod
    def _process_created_entity(cls, entity, **data):
        """
        processes created entity if required.

        it must change the attributes of given entity in-place.

        :param pyrin.database.model.base.BaseEntity entity: created entity.

        :keyword **data: all data that has been passed to create method.
        """
        pass

    @classmethod
    def _create(cls, **data):
        """
        creates an entity with given data.

        it's preferred for this method to return the pk of created entity
        if it is not a composite primary key. this lets the client to fill
        fk fields automatically after create.

        :keyword **data: all data to be passed to create method.

        :rtype: object
        """

        entity = cls.entity(**data)
        cls._process_created_entity(entity, **data)
        entity.save()
        if cls._has_single_primary_key():
            return entity.primary_key()

        return None

    @classmethod
    def _process_updated_entity(cls, entity, **data):
        """
        processes updated entity if required.

        it must change the attributes of given entity in-place.

        :param pyrin.database.model.base.BaseEntity entity: updated entity.

        :keyword **data: all data that has been passed to update method.
        """
        pass

    @classmethod
    def _get(cls, pk):
        """
        gets an entity with given primary key.

        :param object pk: primary key of entity to be get.

        :raises EntityNotFoundError: entity not found error.

        :rtype: pyrin.database.model.base.BaseEntity
        """

        store = get_current_store()
        entity = store.query(cls.entity).get(pk)
        if entity is None:
            raise EntityNotFoundError(_('{name} with primary key [{pk}] does not exist.')
                                      .format(name=cls.name, pk=pk))

        return entity

    @classmethod
    def _update(cls, pk, **data):
        """
        updates an entity with given data.

        :param object pk: entity primary key to be updated.

        :keyword **data: all data to be passed to update method.

        :raises EntityNotFoundError: entity not found error.
        """

        entity = cls._get(pk)
        entity.update(**data)
        cls._process_updated_entity(entity, **data)

    @classmethod
    def _remove(cls, pk):
        """
        deletes an entity with given pk.

        :param object pk: entity primary key to be deleted.
        """

        store = get_current_store()
        pk_name = cls._get_primary_key_name()
        pk_column = cls.entity.get_attribute(pk_name)
        store.query(cls.entity).filter(pk_column == pk).delete()

    def _remove_bulk(self, *pk):
        """
        deletes all entities with given primary keys.

        :param object pk: entity primary key to be deleted.
        """

        store = get_current_store()
        pk_name = self._get_primary_key_name()
        pk_column = self.entity.get_attribute(pk_name)
        store.query(self.entity).filter(pk_column.in_(pk)).delete()

    def _remove_all(self):
        """
        deletes all entities.
        """

        store = get_current_store()
        store.query(self.entity).delete()

    @fast_cache
    def _get_list_entities(self):
        """
        gets all entities that are involved in list select query.

        :rtype: tuple[type[BaseEntity]]
        """

        selectable_fields = self._get_selectable_fields()
        entities = []
        for item in selectable_fields:
            if isinstance(item, InstrumentedAttribute):
                entities.append(item.class_)

        return tuple(set(entities))

    @fast_cache
    def _get_list_labels(self):
        """
        gets all labels that are involved in list select query.

        :rtype: tuple[str]
        """

        selectable_fields = self._get_selectable_fields()
        labels = []
        for item in selectable_fields:
            if isinstance(item, Label) and item.key != self._get_hidden_pk_name():
                labels.append(item.key)

        return tuple(set(labels))

    @fast_cache
    def _get_sortable_fields(self):
        """
        gets all field names that could be used for result ordering by client.

        :rtype: tuple[str]
        """

        list_fields = self._get_list_fields()
        results = []
        for item in list_fields:
            if isinstance(item, (InstrumentedAttribute, Label)) \
                    and item.key != self._get_hidden_pk_name():
                results.append(item.key)

        return tuple(set(results))

    def _get_extra_data_fields(self, for_update):
        """
        gets all extra data fields of this admin page for specified operation.

        :param bool for_update: specifies that fields must be returned for which operation.

        :rtype: tuple[dict]
        """

        fields = []
        if self.extra_data_fields:
            for name, type_ in self.extra_data_fields.items():
                item = dict(field=name, is_fk=False, is_pk=False,
                            title=self._get_field_title(name),
                            form_field_type=validator_services.get_form_field_type(type_))
                validator = validator_services.try_get_validator(self.entity, name)
                if validator is not None:
                    item.update(validator.get_info(for_update))

                item.setdefault('required', not for_update)
                fields.append(item)

        return tuple(fields)

    def _get_data_field(self, name, for_update, primary_keys, foreign_keys):
        """
        gets info of given data field for specified operation.

        :param str name: field name.
        :param bool for_update: specifies that fields must be returned for which operation.
        :param tuple[str] primary_keys: all primary key names.
        :param tuple[str] foreign_keys: all foreign key names.

        :rtype: dict
        """

        is_fk = name in foreign_keys
        info = dict(field=name, title=self._get_field_title(name),
                    is_fk=is_fk, is_pk=name in primary_keys)

        column = self.entity.get_attribute(name)
        if is_fk:
            fk_info = self._get_fk_info(column)
            info.update(fk_info)

        validator = validator_services.try_get_validator(self.entity, column)
        if validator is not None:
            info.update(validator.get_info(for_update))

        return info

    def _validate_duplicate_data_fields(self, writable_fields, readable_fields=None):
        """
        validates that provided field names do not have overlap with extra data field names.

        :param tuple[str] writable_fields: writable field names.
        :param tuple[str] readable_fields: readable field names.

        :raises ExtraDataFieldsAndEntityFieldsOverlapError: extra data fields and
                                                            entity fields overlap error.
        """

        if not self.extra_data_fields:
            return

        fields = set(writable_fields).union(set(readable_fields or []))
        extra_fields = set(self.extra_data_fields)
        duplicates = fields.intersection(extra_fields)
        if duplicates:
            raise ExtraDataFieldsAndEntityFieldsOverlapError('These extra data fields have the '
                                                             'same name with entity [{entity}] '
                                                             'fields. please rename these extra '
                                                             'data fields in admin [{admin}] '
                                                             'class: {fields}'
                                                             .format(entity=self.entity,
                                                                     admin=self,
                                                                     fields=list(duplicates)))

    def _get_data_fields(self, writable_fields, for_update,
                         primary_keys, foreign_keys,
                         readable_fields=None):
        """
        gets all data fields of this admin page for specified operation.

        :param tuple[str] writable_fields: writable field names.
        :param bool for_update: specifies that fields must be returned for which operation.
        :param tuple[str] primary_keys: all primary key names.
        :param tuple[str] foreign_keys: all foreign key names.
        :param tuple[str] readable_fields: readable field names.

        :rtype: tuple[dict]
        """

        self._validate_duplicate_data_fields(writable_fields, readable_fields)
        fields = []
        for name in writable_fields:
            item = self._get_data_field(name, for_update, primary_keys, foreign_keys)
            item.setdefault('required', False)
            fields.append(item)

        extra_fields = self._get_extra_data_fields(for_update)
        fields.extend(extra_fields)
        fields = dict_utils.extended_sort(fields, 'required', reverse=True)

        if readable_fields:
            for name in readable_fields:
                if name not in writable_fields and name not in primary_keys:
                    item = self._get_data_field(name, for_update, primary_keys, foreign_keys)
                    item.update(read_only=True)
                    fields.append(item)

        return tuple(fields)

    @fast_cache
    def _get_create_fields(self):
        """
        gets all fields which should be used for create page.

        :rtype: tuple[dict]
        """

        writable_columns = self.entity.writable_primary_key_columns + \
            self.entity.writable_foreign_key_columns + self.entity.writable_columns

        return self._get_data_fields(writable_columns, False,
                                     self.entity.primary_key_columns,
                                     self.entity.foreign_key_columns)

    @fast_cache
    def _get_create_fields_dict(self):
        """
        gets all fields which should be used for create page as a dict.

        :rtype: dict
        """

        result = dict()
        fields = self._get_create_fields()
        for item in fields:
            result[item.get('field')] = item

        return result

    @fast_cache
    def _get_update_fields(self):
        """
        gets all fields which should be used for update page.

        :rtype: tuple[dict]
        """

        writable_columns = self.entity.writable_primary_key_columns + \
            self.entity.writable_foreign_key_columns + self.entity.writable_columns

        readable_columns = self.entity.readable_primary_key_columns + \
            self.entity.readable_foreign_key_columns + self.entity.readable_columns

        return self._get_data_fields(writable_columns, True,
                                     self.entity.primary_key_columns,
                                     self.entity.foreign_key_columns,
                                     readable_columns)

    @fast_cache
    def _get_update_fields_dict(self):
        """
        gets all fields which should be used for update page as a dict.

        :rtype: dict
        """

        result = dict()
        fields = self._get_update_fields()
        for item in fields:
            result[item.get('field')] = item

        return result

    def _get_link_info(self, title, register_name, type_=None,
                       button_type=None, new_tab=True, **filters):
        """
        gets the required link info for given inputs.

        the result of this method must be returned by admin methods
        which want to render a link on client list view.
        the aforementioned methods must be decorated using `@link` decorator.

        :param str title: the title of the link.

        :param str | type[BaseEntity] register_name: the register name or the entity class of
                                                     the related admin class to view its list
                                                     page after clicking on this link.

        :param str type_: the link type.
                          if not provided defaults to `button`.
        :enum type_:
            BUTTON = 'button'
            LINK = 'link'

        :param str button_type: the link button type.
                                this is only used if the `type_` is `button`.
                                defaults to `outlined` if not provided.
        :enum button_type:
            CONTAINED = 'contained'
            OUTLINED = 'outlined'
            TEXT = 'text'

        :param bool new_tab: open this link in a new tab.
                             defaults to True if not provided.

        :param **filters: any query strings that must be passed to the related
                          page after clicking on this link.

        :raises AdminPageNotFoundError: admin page not found error.

        :returns: dict(str title,
                       str register_name,
                       str type,
                       str button_type,
                       bool new_tab,
                       dict filters)
        :rtype: dict
        """

        if not isinstance(register_name, str):
            register_name = admin_services.register_name_for(register_name)

        if type_ not in LinkTypeEnum:
            type_ = LinkTypeEnum.BUTTON

        if button_type not in ButtonTypeEnum:
            button_type = ButtonTypeEnum.OUTLINED

        return dict(title=title,
                    register_name=register_name,
                    type=type_,
                    button_type=button_type,
                    new_tab=new_tab,
                    filters=filters)

    @fast_cache
    def _get_common_metadata(self):
        """
        gets the common metadata of this admin page.

        :rtype: dict
        """

        metadata = dict()
        metadata['name'] = self.name
        metadata['plural_name'] = self.get_plural_name()
        metadata['register_name'] = self.get_register_name()
        metadata['category'] = self.get_category()
        metadata['configs'] = admin_services.get_configs()
        return metadata

    def get_entity(self):
        """
        gets the entity class of this admin page.

        :rtype: type[BaseEntity]
        """

        return self.entity

    @fast_cache
    def get_register_name(self):
        """
        gets the register name of this admin page.

        :rtype: str
        """

        return self.register_name.lower()

    @fast_cache
    def get_category(self):
        """
        gets the category of this admin page.

        :rtype: str
        """

        if self.category not in (None, '') and not self.category.isspace():
            return self.category.upper()

        package = path_utils.get_object_package_name(self)
        if package is not None:
            package = package.replace('_', ' ')
            return package.upper()

        return admin_services.get_default_category()

    @fast_cache
    def get_plural_name(self):
        """
        gets the plural name of this admin page.

        :rtype: str
        """

        return self.plural_name or string_utils.pluralize(self.name)

    def get_singular_name(self):
        """
        gets the singular name of this admin page.

        :rtype: str
        """

        return self.name

    def get(self, pk):
        """
        gets an entity with given primary key.

        :param object pk: primary key of entity to be get.

        :raises EntityNotFoundError: entity not found error.

        :rtype: pyrin.database.model.base.BaseEntity
        """

        validator_services.validate(self.entity, **self._get_primary_key_holder(pk))
        return self._get(pk)

    def find(self, **filters):
        """
        finds entities with given filters.

        :keyword **filters: all filters to be passed to related find service.

        :raises ListFieldRequiredError: list field required error.

        :rtype: list[ROW_RESULT]
        """

        self._validate_filters(filters)
        store = get_current_store()
        query = store.query(*self._get_selectable_fields())
        query = self._perform_joins(query)
        query = self._filter_query(query, **filters)
        query = self._perform_order_by(query, **filters)
        query = self._paginate_query(query, **filters)
        results = query.all()
        return self._process_find_results(results, **filters)

    @classmethod
    def create(cls, **data):
        """
        creates an entity with given data.

        it's preferred for this method to return the pk of created entity
        if it is not a composite primary key. this lets the client to fill
        fk fields automatically after create.

        :keyword **data: all data to be passed to related create service.

        :rtype: object
        """

        validator_services.validate_dict(cls.entity, data)
        cls._validate_extra_fields(data)
        if cls.create_service is not None:
            result = cls.create_service(**data)
            if cls._has_single_primary_key():
                return result
            return None
        else:
            return cls._create(**data)

    @classmethod
    def update(cls, pk, **data):
        """
        updates an entity with given data.

        :param object pk: entity primary key to be updated.

        :keyword **data: all data to be passed to related update service.

        :raises EntityNotFoundError: entity not found error.
        """

        validator_services.validate(cls.entity, **cls._get_primary_key_holder(pk))
        validator_services.validate_dict(cls.entity, data, for_update=True)
        if cls.update_service is not None:
            cls.update_service(pk, **data)
        else:
            cls._update(pk, **data)

    @classmethod
    def remove(cls, pk):
        """
        deletes an entity with given pk.

        :param object pk: entity primary key to be deleted.
        """

        validator_services.validate(cls.entity, **cls._get_primary_key_holder(pk))
        if cls.remove_service is not None:
            cls.remove_service(pk)
        else:
            cls._remove(pk)

    def remove_bulk(self, pk):
        """
        deletes entities with given primary keys.

        :param object | list[object] pk: entity primary keys to be deleted.
        """

        pk = misc_utils.make_iterable(pk)
        for item in pk:
            validator_services.validate(self.entity, **self._get_primary_key_holder(item))

        self._remove_bulk(*pk)

    def remove_all(self):
        """
        deletes all entities.
        """

        self._remove_all()

    def call_method(self, name, argument):
        """
        calls the method with given name with given argument and returns the result.

        :param str name: method name.
        :param ROW_RESULT argument: the method argument.

        :raises InvalidMethodNameError: invalid method name error.

        :rtype: object
        """

        if name not in self.method_names:
            raise InvalidMethodNameError('Method [{method}] is not present in [{admin}] class.'
                                         .format(method=name, admin=self))

        method = getattr(self, name)
        return method(argument)

    def has_get_permission(self):
        """
        gets a value indicating that this admin page has get permission.

        note that entities with composite primary key does not support get.

        :rtype: bool
        """

        return self._has_single_primary_key() and self.get_permission

    def has_create_permission(self):
        """
        gets a value indicating that this admin page has create permission.

        :rtype: bool
        """

        return self.create_permission

    def has_update_permission(self):
        """
        gets a value indicating that this admin page has update permission.

        note that entities with composite primary key does not support update.

        :rtype: bool
        """

        return self._has_single_primary_key() and self.update_permission

    def has_remove_permission(self):
        """
        gets a value indicating that this admin page has single or bulk remove permission.

        note that entities with composite primary key does not support single or bulk remove.

        :rtype: bool
        """

        return self._has_single_primary_key() and self.remove_permission

    def has_remove_all_permission(self):
        """
        gets a value indicating that this admin page has remove all permission.

        :rtype: bool
        """

        return self.remove_all_permission

    @fast_cache
    def get_main_metadata(self):
        """
        gets the main metadata of this admin page.

        :rtype: dict
        """

        metadata = dict()
        metadata.update(self._get_common_metadata())
        metadata['has_create_permission'] = self.has_create_permission()
        return metadata

    @fast_cache
    def get_find_metadata(self):
        """
        gets the find metadata of this admin page.

        :rtype: dict
        """

        metadata = dict()
        metadata.update(self._get_common_metadata())
        metadata['datasource_info'] = self._get_list_datasource_info()
        metadata['field_names'] = self._get_list_field_names()
        metadata['sortable_fields'] = self._get_sortable_fields()
        metadata['has_create_permission'] = self.has_create_permission()
        metadata['has_remove_permission'] = self.has_remove_permission()
        metadata['has_remove_all_permission'] = self.has_remove_all_permission()
        metadata['has_get_permission'] = self.has_get_permission()
        metadata['paged'] = self.list_paged
        metadata['page_size'] = self._get_page_size()
        metadata['max_page_size'] = self._get_max_page_size()
        metadata['page_size_options'] = self._get_page_size_options()
        metadata['table_type'] = self.list_table_type
        metadata['pagination_type'] = self.list_pagination_type
        metadata['pagination_position'] = self.list_pagination_position
        metadata['enable_export'] = self.list_enable_export
        metadata['export_name'] = self._get_list_export_name()
        metadata['link_pk'] = self.list_link_pk
        metadata['link_fk'] = self.list_link_fk
        metadata['column_selection'] = self.list_column_selection
        metadata['column_ordering'] = self.list_column_ordering
        metadata['grouping'] = self.list_grouping
        metadata['search'] = self.list_search
        metadata['search_debounce_interval'] = self.list_search_debounce_interval
        metadata['datetime_format'] = self.list_datetime_format
        metadata['date_format'] = self.list_date_format
        metadata['time_format'] = self.list_time_format
        metadata['locale'] = self.list_locale
        metadata['max_body_height'] = self.list_max_body_height
        return metadata

    @fast_cache
    def get_create_metadata(self):
        """
        gets the create metadata of this admin page.

        :rtype: dict
        """

        metadata = dict()
        metadata.update(self._get_common_metadata())
        metadata['has_create_permission'] = self.has_create_permission()
        metadata['data_fields'] = self._get_create_fields()
        metadata['data_fields_dict'] = self._get_create_fields_dict()
        return metadata

    @fast_cache
    def get_update_metadata(self):
        """
        gets the update metadata of this admin page.

        :rtype: dict
        """

        metadata = dict()
        metadata.update(self._get_common_metadata())
        metadata['has_update_permission'] = self.has_update_permission()
        metadata['has_get_permission'] = self.has_get_permission()
        metadata['has_remove_permission'] = self.has_remove_permission()
        metadata['data_fields'] = self._get_update_fields()
        metadata['data_fields_dict'] = self._get_update_fields_dict()
        return metadata

    def populate_caches(self):
        """
        populates required caches of this admin page.
        """

        self.get_main_metadata()
        self.get_find_metadata()
        self.get_create_metadata()
        self.get_update_metadata()
        self._get_list_search_fields_to_column_map()
        self._get_primary_keys()
        self._get_default_list_fields()
        self._get_list_entities()
        self._get_list_labels()

    @property
    def method_names(self):
        """
        gets the list of all method names of this admin page to be used for result processing.

        :rtype: tuple[str]
        """

        return self._method_names



"""
admin page decorators module.
"""


def link(method):
    """
    decorator to flag an admin method as rendering a link on client list view.

    :param object method: method to be flagged.

    :returns: method.
    :rtype: type
    """

    setattr(method, 'is_link', True)
    return method



# -*- coding: utf-8 -*-
"""
audit security authorizers module.
"""

from pyrin.core.globals import _
from pyrin.security.authorization.decorators import authorizer
from pyrin.security.enumerations import InternalAuthenticatorEnum
from pyrin.security.authorization.handlers.internal import InternalAuthorizer
from pyrin.audit.security.exceptions import AuditAccessNotAllowedError


@authorizer()
class AuditAuthorizer(InternalAuthorizer):
    """
    audit authorizer class.
    """

    _name = InternalAuthenticatorEnum.AUDIT

    def _authorize_access(self, user, **options):
        """
        authorizes the given internal user for audit access.

        :param int user: internal user id to be checked if it is authorized.

        :keyword dict user_info: internal user info to be used to check for user.

        :raises AuditAccessNotAllowedError: audit access not allowed error.
        """

        user_info = options.get('user_info')
        if not user_info or user_info.get('audit_access') is not True:
            raise AuditAccessNotAllowedError(_('You are not allowed to access the audit api. '
                                               'If you think that this is a mistake, please '
                                               'contact the support team.'))



# -*- coding: utf-8 -*-
"""
authorization handlers internal module.
"""

from pyrin.security.authorization.handlers.base import AuthorizerBase


class InternalAuthorizer(AuthorizerBase):
    """
    internal authorizer class.

    all application internal authorizers must be subclassed from this.
    for example admin, swagger or audit authorizers.
    """

    def _is_superuser(self, user, **options):
        """
        gets a value indicating that given internal user is superuser.

        :param int user: internal user id to be checked if it is superuser.

        :keyword dict user_info: internal user info to be used to check for user.

        :rtype: bool
        """

        user_info = options.get('user_info')
        return user_info and user_info.get('is_superuser') is True

    def _is_active(self, user, **options):
        """
        gets a value indicating that the given internal user is active.

        :param int user: internal user id to be checked if it is active.

        :keyword dict user_info: internal user info to be used to check for user.

        :rtype: bool
        """

        user_info = options.get('user_info')
        return user_info and user_info.get('is_active') is True



# -*- coding: utf-8 -*-
"""
swagger base module.
"""

import re
import os

from uuid import UUID
from copy import deepcopy
from collections import defaultdict

from flasgger import Swagger
from flasgger.marshmallow_apispec import SwaggerView, convert_schemas
from flasgger.constants import OPTIONAL_FIELDS, OPTIONAL_OAS3_FIELDS
from flasgger.utils import extract_definitions, parse_definition_docstring, \
    get_vendor_extension_fields, merge_specs, parse_docstring, is_valid_method_view, \
    has_valid_dispatch_view_docs

import pyrin.database.services as database_services
import pyrin.api.swagger.services as swagger_services
import pyrin.configuration.services as config_services
import pyrin.database.paging.services as paging_services
import pyrin.application.services as application_services
import pyrin.globalization.locale.services as locale_services
import pyrin.globalization.datetime.services as datetime_services
import pyrin.processor.response.status.services as status_services

from pyrin.api.router.handlers.protected import ProtectedRoute
from pyrin.core.enumerations import HTTPMethodEnum, ClientErrorResponseCodeEnum
from pyrin.api.swagger.enumerations import ParameterAttributeEnum, ParameterLocationEnum, \
    ParameterTypeEnum, DocstringSectionEnum, ParameterFormatEnum


class ExtendedSwagger(Swagger):
    """
    extended swagger class.

    this class extends `Swagger` class to enable adding common
    schema attributes into registered specs.
    """

    DEFAULT_CONFIG = {
        'headers': [
        ],
        'specs': [
            {
                'endpoint': 'swagger',
                'route': '/swagger.json',
                'rule_filter': lambda rule: True,  # all in
                'model_filter': lambda tag: True,  # all in
            }
        ],
        'static_url_path': '/flasgger_static',
        'swagger_ui': True,
        'specs_route': '/swagger/'
    }

    def _get_parameters_section(self, swag):
        """
        gets parameters section from given swag info.

        :param dict swag: swag info.

        :rtype: list[dict]
        """

        return swag.setdefault(DocstringSectionEnum.PARAMETERS, [])

    def _get_responses_section(self, swag):
        """
        gets responses section from given swag info.

        :param dict swag: swag info.

        :rtype: dict
        """

        return swag.setdefault(DocstringSectionEnum.RESPONSES, {})

    def _get_security_section(self, swag):
        """
        gets security section from given swag info.

        :param dict swag: swag info.

        :rtype: list[dict]
        """

        return swag.setdefault(DocstringSectionEnum.SECURITY, [])

    def _get_tags_section(self, swag):
        """
        gets tags section from given swag info.

        :param dict swag: swag info.

        :rtype: list[str]
        """

        return swag.setdefault(DocstringSectionEnum.TAGS, [])

    def _get_parameter(self, name, parameters):
        """
        gets the parameter with given name from provided params.

        it may return None if it does not exist.

        :param str name: parameter name.
        :param list[dict] parameters: list of parameters.

        :rtype: dict
        """

        for param in parameters:
            if param.get(ParameterAttributeEnum.NAME) == name:
                return param

        return None

    def _add_or_update_parameter(self, parameters, name, type_, in_, **options):
        """
        adds a new parameter into given parameters or updates the existing one.

        :param list[dict] parameters: list of parameters.
        :param str name: new parameter name.
        :param str type_: new parameter type.
        :enum type_:
            INTEGER = 'integer'
            NUMBER = 'number'
            BOOLEAN = 'boolean'
            STRING = 'string'
            ARRAY = 'array'
            OBJECT = 'object'

        :param in_: new parameter place.
        :enum in_:
            PATH = 'path'
            QUERY = 'query'
            BODY = 'body'
            HEADER = 'header'
            FORM = 'formData'

        :keyword bool required: specifies that this parameter is required.
                                defaults to False if not provided.

        :keyword str description: new parameter description.
                                  defaults to an empty string if not provided.

        :keyword str format: format to be used for given type.
        :enum format:
            UUID = 'uuid'
            EMAIL = 'email'
            DATE = 'date'
            DATE_TIME = 'date-time'
            PASSWORD = 'password'
            BYTE = 'byte'
            URI = 'uri'
            HOSTNAME = 'hostname'
            IPV4 = 'ipv4'
            IPV6 = 'ipv6'
            DOUBLE = 'double'
            FLOAT = 'float'

        :keyword bool add_first: add the new parameter at the beginning of
                                 available params. defaults to False if not
                                 provided and the new param will be appended
                                 to the end.
        """

        param = self._get_parameter(name, parameters)
        is_new = param is None
        add_first = options.get('add_first', False)
        format_ = options.get('format')
        required = options.get('required', False)
        description = options.get('description')
        new_param = dict()
        if not is_new:
            new_param = param

        new_param.setdefault(ParameterAttributeEnum.NAME, name)
        new_param.setdefault(ParameterAttributeEnum.REQUIRED, required)
        new_param.setdefault(ParameterAttributeEnum.IN, in_)

        if type_ is not None:
            new_param.setdefault(ParameterAttributeEnum.TYPE, type_)

        if format_ is not None:
            new_param.setdefault(ParameterAttributeEnum.FORMAT, format_)

        if description is not None:
            new_param.setdefault(ParameterAttributeEnum.DESCRIPTION, description)

        if is_new:
            if add_first is True:
                parameters.insert(0, new_param)
            else:
                parameters.append(new_param)

    def _get_param_location(self, name, rule, verb):
        """
        gets the location that this parameter must be placed in a request.

        if the parameter is from url params of given rule, it will return `path`.
        if the http method is one of `GET`, `OPTIONS` or `HEAD` it will return `query`.
        otherwise it will return `formData`.

        :param st name: parameter name.
        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this parameter.
        :param str verb: http method name.

        :rtype: str
        """

        if name in rule.arguments:
            return ParameterLocationEnum.PATH

        if verb.upper() in (HTTPMethodEnum.HEAD,
                            HTTPMethodEnum.GET,
                            HTTPMethodEnum.OPTIONS):
            return ParameterLocationEnum.QUERY

        return ParameterLocationEnum.FORM

    def _add_or_update_response(self, responses, rule, code, **options):
        """
        adds a new response into given responses or updates the existing one.

        :param dict responses: dict of responses.
        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this response.
        :param int code: response status code.

        :keyword str description: new response description.
                                  defaults to a generic message if not provided.
        """

        response = responses.get(code)
        if response is None:
            response = dict()

        description = options.get('description')
        if description in (None, ''):
            description = self._get_response_description(code)

        if description is not None:
            response.setdefault(ParameterAttributeEnum.DESCRIPTION, description)

        self._wrap_schema(response, rule, code)
        responses[code] = response

    def _get_response_description(self, code):
        """
        gets a generic response description for given status code.

        it may return None if provided code is unknown.

        :param int code: status code.

        :rtype: str
        """

        if code == ClientErrorResponseCodeEnum.UNPROCESSABLE_ENTITY:
            return 'a business error such as validation error occurred'

        if code == ClientErrorResponseCodeEnum.UNAUTHORIZED:
            return 'user has not been authenticated'

        if code == ClientErrorResponseCodeEnum.FORBIDDEN:
            return 'you do not have the required permissions to access this resource'

        if code == ClientErrorResponseCodeEnum.BAD_REQUEST:
            return 'required parameters are not provided'

        if code == ClientErrorResponseCodeEnum.NOT_FOUND:
            return 'the requested resource or entity not found'

        if status_services.is_success(code, strict_status=False):
            return 'successful execution of service'

        if status_services.is_redirection(code, strict_status=False):
            return 'redirection is required'

        if status_services.is_information(code, strict_status=False):
            return 'informational response'

        if status_services.is_client_error(code, strict_status=False):
            return 'a client error has been occurred'

        if status_services.is_server_error(code, strict_status=False):
            return 'an internal server error has been occurred'

        return None

    def _wrap_schema(self, response, rule, code):
        """
        wraps the given response's schema into a dict.

        if the available schema is already a dict, it does nothing.

        :param dict response: response dict.
        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this response.
        :param int code: response status code.
        """

        current_schema = response.get(ParameterAttributeEnum.SCHEMA)
        if current_schema is None:
            current_schema = dict()

        type_ = current_schema.get(ParameterAttributeEnum.TYPE)
        if type_ == ParameterTypeEnum.OBJECT or \
                ParameterAttributeEnum.PROPERTIES in current_schema:
            current_schema[ParameterAttributeEnum.TYPE] = ParameterTypeEnum.OBJECT
            return

        if ParameterAttributeEnum.ITEMS in current_schema:
            type_ = ParameterTypeEnum.ARRAY

        if status_services.is_error(code, strict_status=False):
            current_schema = self._get_error_schema()
        elif type_ == ParameterTypeEnum.ARRAY:
            if rule.is_paged:
                current_schema = self._get_paged_list_schema(current_schema)
            else:
                current_schema = self._get_list_schema(current_schema)
        elif type_ in (ParameterTypeEnum.INTEGER, ParameterTypeEnum.NUMBER,
                       ParameterTypeEnum.STRING, ParameterTypeEnum.BOOLEAN):
            current_schema = self._get_single_value_schema(current_schema)

        if len(current_schema) > 0:
            response[ParameterAttributeEnum.SCHEMA] = current_schema

    def _get_error_schema(self):
        """
        gets the error schema.

        :rtype: dict
        """

        schema = dict()
        properties = dict()
        schema[ParameterAttributeEnum.TYPE] = ParameterTypeEnum.OBJECT
        schema[ParameterAttributeEnum.PROPERTIES] = properties

        self._add_or_update_property(properties, 'code',
                                     ParameterTypeEnum.INTEGER,
                                     description='error status code')

        self._add_or_update_property(properties, 'data',
                                     ParameterTypeEnum.OBJECT,
                                     description='error extended data')

        self._add_or_update_property(properties, 'request_id',
                                     ParameterTypeEnum.STRING,
                                     format=ParameterFormatEnum.UUID,
                                     description='current request id')

        self._add_or_update_property(properties, 'message',
                                     ParameterTypeEnum.STRING,
                                     description='error message')

        return schema

    def _get_paged_list_schema(self, current_schema):
        """
        gets a paged list response schema.

        :param dict current_schema: current schema to be processed.

        :rtype: dict
        """

        results_schema = current_schema.get(ParameterAttributeEnum.ITEMS)
        if results_schema is None:
            results_schema = dict()

        results_schema.setdefault(ParameterAttributeEnum.TYPE, ParameterTypeEnum.STRING)

        schema = dict()
        properties = dict()
        schema[ParameterAttributeEnum.TYPE] = ParameterTypeEnum.OBJECT
        schema[ParameterAttributeEnum.PROPERTIES] = properties

        self._add_or_update_property(properties, 'count',
                                     ParameterTypeEnum.INTEGER,
                                     description='results count of the current page')

        self._add_or_update_property(properties, 'count_total',
                                     ParameterTypeEnum.INTEGER,
                                     description='results count of all pages')

        self._add_or_update_property(properties, 'next',
                                     ParameterTypeEnum.STRING,
                                     format=ParameterFormatEnum.URI,
                                     description='next page url')

        self._add_or_update_property(properties, 'previous',
                                     ParameterTypeEnum.STRING,
                                     format=ParameterFormatEnum.URI,
                                     description='previous page url')

        self._add_or_update_property(properties, 'results',
                                     ParameterTypeEnum.ARRAY,
                                     description='result items',
                                     schema=results_schema)

        return schema

    def _get_list_schema(self, current_schema):
        """
        gets a list response schema.

        :param dict current_schema: current schema to be processed.

        :rtype: dict
        """

        results_schema = current_schema.get(ParameterAttributeEnum.ITEMS)
        if results_schema is None:
            results_schema = dict()

        results_schema.setdefault(ParameterAttributeEnum.TYPE, ParameterTypeEnum.STRING)

        schema = dict()
        properties = dict()
        schema[ParameterAttributeEnum.TYPE] = ParameterTypeEnum.OBJECT
        schema[ParameterAttributeEnum.PROPERTIES] = properties

        self._add_or_update_property(properties, 'count_total',
                                     ParameterTypeEnum.INTEGER,
                                     description='results count')

        self._add_or_update_property(properties, 'results',
                                     ParameterTypeEnum.ARRAY,
                                     description='result items',
                                     schema=results_schema)

        return schema

    def _get_single_value_schema(self, current_schema):
        """
        gets a single value response schema.

        :param dict current_schema: current schema to be processed.

        :rtype: dict
        """

        description = current_schema.get(ParameterAttributeEnum.DESCRIPTION)
        type_ = current_schema.get(ParameterAttributeEnum.TYPE)
        if type_ is None:
            type_ = ParameterTypeEnum.STRING

        schema = dict()
        properties = dict()
        schema[ParameterAttributeEnum.TYPE] = ParameterTypeEnum.OBJECT
        schema[ParameterAttributeEnum.PROPERTIES] = properties

        self._add_or_update_property(properties, 'value', type_,
                                     description=description)

        return schema

    def _add_or_update_property(self, properties, name, type_, **options):
        """
        adds a new property into given properties or updates the existing one.

        :param dict properties: dict of properties.
        :param str name: property name.
        :param str type_: property type.
        :enum type_:
            INTEGER = 'integer'
            NUMBER = 'number'
            BOOLEAN = 'boolean'
            STRING = 'string'
            ARRAY = 'array'
            OBJECT = 'object'

        :keyword str description: property description.
        :keyword str format: property type format.
        :enum format:
            UUID = 'uuid'
            EMAIL = 'email'
            DATE = 'date'
            DATE_TIME = 'date-time'
            PASSWORD = 'password'
            BYTE = 'byte'
            URI = 'uri'
            HOSTNAME = 'hostname'
            IPV4 = 'ipv4'
            IPV6 = 'ipv6'
            DOUBLE = 'double'
            FLOAT = 'float'

        :keyword dict schema: schema of the property.
                              it will be used when type is `array` or `dict`.
        """

        property_ = properties.get(name)
        if property_ is None:
            property_ = dict()

        if type_ is None:
            type_ = ParameterTypeEnum.STRING

        property_.setdefault(ParameterAttributeEnum.TYPE, type_)
        schema = options.get('schema')
        description = options.get('description')
        format_ = options.get('format')
        if description not in (None, ''):
            property_.setdefault(ParameterAttributeEnum.DESCRIPTION, description)

        if format_ not in (None, ''):
            property_.setdefault(ParameterAttributeEnum.FORMAT, format_)

        if schema not in (None, ''):
            if type_ == ParameterTypeEnum.ARRAY:
                property_[ParameterAttributeEnum.ITEMS] = schema
            elif type_ == ParameterTypeEnum.OBJECT:
                property_[ParameterAttributeEnum.PROPERTIES] = schema

        properties[name] = property_

    def _fix_metadata(self, rule, verb, swag):
        """
        fixes metadata of given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        self._add_or_fix_required_parameters(rule, verb, swag)
        self._fix_optional_parameters(rule, verb, swag)
        self._add_paging_parameters(rule, verb, swag)
        self._add_order_by_parameter(rule, verb, swag)
        self._add_locale_parameter(rule, verb, swag)
        self._add_timezone_parameter(rule, verb, swag)
        self._add_security_definitions(rule, verb, swag)
        self._fix_responses(rule, verb, swag)
        self._add_authentication_failed_response(rule, verb, swag)
        self._add_business_error_response(rule, verb, swag)
        self._add_permission_denied_response(rule, verb, swag)
        self._add_successful_response(rule, verb, swag)
        self._add_bad_request_response(rule, verb, swag)
        self._fix_parameter_types(rule, verb, swag)
        self._add_tags(rule, verb, swag)

    def _add_locale_parameter(self, rule, verb, swag):
        """
        adds locale parameter into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        params = self._get_parameters_section(swag)
        self._add_or_update_parameter(params,
                                      locale_services.get_locale_key(),
                                      ParameterTypeEnum.STRING,
                                      ParameterLocationEnum.QUERY,
                                      required=False,
                                      description='request locale to be sent. '
                                                  'for example en, fa, fr ...')

    def _add_timezone_parameter(self, rule, verb, swag):
        """
        adds timezone parameter into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        params = self._get_parameters_section(swag)
        self._add_or_update_parameter(params,
                                      datetime_services.get_timezone_key(),
                                      ParameterTypeEnum.STRING,
                                      ParameterLocationEnum.QUERY,
                                      required=False,
                                      description='request timezone to be sent. '
                                                  'for example UTC, Asia/Tehran ...')

    def _add_paging_parameters(self, rule, verb, swag):
        """
        adds paging parameters into given swag info if provided rule is paged.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        if rule.is_paged is True:
            page, page_size = paging_services.get_paging_param_names()
            params = self._get_parameters_section(swag)
            self._add_or_update_parameter(params, page,
                                          ParameterTypeEnum.INTEGER,
                                          ParameterLocationEnum.QUERY,
                                          required=False, description='page number to be get')
            self._add_or_update_parameter(params, page_size,
                                          ParameterTypeEnum.INTEGER,
                                          ParameterLocationEnum.QUERY,
                                          required=False, description='page size to be get')

    def _add_order_by_parameter(self, rule, verb, swag):
        """
        adds order by parameter into given swag info if the rule is ordered.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        if rule.ordered is True:
            params = self._get_parameters_section(swag)
            self._add_or_update_parameter(params,
                                          database_services.get_ordering_key(),
                                          ParameterTypeEnum.STRING,
                                          ParameterLocationEnum.QUERY,
                                          required=False,
                                          description='column name or list of column names '
                                                      'to be used for result ordering')

    def _add_or_fix_required_parameters(self, rule, verb, swag):
        """
        adds or fixes required parameters into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        params = self._get_parameters_section(swag)
        required_args = rule.required_arguments.union(rule.arguments)
        for name in required_args:
            type_ = None
            format_ = None
            in_ = self._get_param_location(name, rule, verb)
            if in_ == ParameterLocationEnum.PATH:
                argument_type = rule.get_argument_type(name)
                if argument_type is int:
                    type_ = ParameterTypeEnum.INTEGER
                elif argument_type is float:
                    type_ = ParameterTypeEnum.NUMBER
                    format_ = ParameterFormatEnum.FLOAT
                elif argument_type is str:
                    type_ = ParameterTypeEnum.STRING
                elif argument_type is UUID:
                    type_ = ParameterTypeEnum.STRING
                    format_ = ParameterFormatEnum.UUID

            self._add_or_update_parameter(params, name, type_, in_,
                                          required=True, add_first=True,
                                          format=format_)

            param = self._get_parameter(name, params)
            if param is not None:
                param[ParameterAttributeEnum.REQUIRED] = True

    def _fix_optional_parameters(self, rule, verb, swag):
        """
        fixes metadata of optional parameters in given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        required_args = rule.required_arguments.union(rule.arguments)
        params = self._get_parameters_section(swag)
        for item in params:
            name = item.get(ParameterAttributeEnum.NAME)
            if name in required_args:
                continue

            item.setdefault(ParameterAttributeEnum.REQUIRED, False)
            in_ = self._get_param_location(name, rule, verb)
            item.setdefault(ParameterAttributeEnum.IN, in_)

    def _fix_parameter_types(self, rule, verb, swag):
        """
        fixes all parameters that have no type and sets their type to `string`.

        this is required to prevent swagger ui client from erroneous behaviors.
        also when a parameter has no type, no validation will be done for it by
        swagger even if the parameter is required.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        params = self._get_parameters_section(swag)
        for item in params:
            item.setdefault(ParameterAttributeEnum.TYPE, ParameterTypeEnum.STRING)

    def _add_security_definitions(self, rule, verb, swag):
        """
        adds security definitions into given swag info if provided rule is protected.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        if isinstance(rule, ProtectedRoute):
            definitions = config_services.get_active('swagger', 'securityDefinitions')
            if definitions is not None and len(definitions) > 0:
                security = self._get_security_section(swag)
                for name in definitions:
                    item = dict()
                    item[name] = []
                    security.append(item)

    def _add_permission_denied_response(self, rule, verb, swag):
        """
        adds permission denied response into given swag info if provided rule is protected.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        if isinstance(rule, ProtectedRoute) and len(rule.permissions) > 0:
            responses = self._get_responses_section(swag)
            self._add_or_update_response(responses, rule,
                                         ClientErrorResponseCodeEnum.FORBIDDEN)

    def _add_authentication_failed_response(self, rule, verb, swag):
        """
        adds authentication failed response into given swag info if provided rule is protected.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        if isinstance(rule, ProtectedRoute):
            responses = self._get_responses_section(swag)
            self._add_or_update_response(responses, rule,
                                         ClientErrorResponseCodeEnum.UNAUTHORIZED)

    def _add_business_error_response(self, rule, verb, swag):
        """
        adds business error response into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        responses = self._get_responses_section(swag)
        self._add_or_update_response(responses, rule,
                                     ClientErrorResponseCodeEnum.UNPROCESSABLE_ENTITY)

    def _add_successful_response(self, rule, verb, swag):
        """
        adds successful response into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        status_code = rule.status_code
        if status_code is None:
            status_code = status_services.get_status_code(method=verb.upper())

        responses = self._get_responses_section(swag)
        self._add_or_update_response(responses, rule, status_code)

    def _add_bad_request_response(self, rule, verb, swag):
        """
        adds bad request response into given swag info if provided rule has required arguments.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        extra_args = rule.required_arguments.difference(rule.arguments)
        if len(extra_args) > 0:
            responses = self._get_responses_section(swag)
            self._add_or_update_response(responses, rule,
                                         ClientErrorResponseCodeEnum.BAD_REQUEST)

    def _fix_responses(self, rule, verb, swag):
        """
        adds required metadata into given swag info for all responses.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        responses = self._get_responses_section(swag)
        for code, response in responses.items():
            if response is None:
                response = dict()

            description = self._get_response_description(code)
            response.setdefault(ParameterAttributeEnum.DESCRIPTION, description)
            self._wrap_schema(response, rule, code)
            responses[code] = response

    def _add_tags(self, rule, verb, swag):
        """
        adds required tags into given swag info.

        :param pyrin.api.router.handlers.base.RouteBase rule: related rule to this swag info.
        :param str verb: http method name.
        :param dict swag: swag info.
        """

        tags = swagger_services.get_tags(rule, verb.upper())
        if len(tags) > 0:
            tag_section = self._get_tags_section(swag)
            tag_section.extend(tags)

    def _should_expose_rule(self, rule):
        """
        gets a value indicating that given rule must be exposed on swagger.

        :param pyrin.api.router.handlers.base.RouteBase rule: rule to be checked.

        :rtype: bool
        """

        return rule.swagger

    def get_apispecs(self, endpoint='swagger'):
        """
        gets api specs for given endpoint.

        we have to duplicate this whole method to be able to hook
        into fetching api specs from docstrings.

        :param str endpoint: endpoint to get its api specs.

        :rtype: dict
        """

        if not self.app.debug and endpoint in self.apispecs:
            return self.apispecs[endpoint]

        spec = None
        for _spec in self.config['specs']:
            if _spec['endpoint'] == endpoint:
                spec = _spec
                break
        if not spec:
            raise RuntimeError('Can`t find specs by endpoint [{endpoint}], '
                               'check your flasgger`s configs.'
                               .format(endpoint=endpoint))

        data = {
            # try to get from config['SWAGGER']['info']
            # then config['SWAGGER']['specs'][x]
            # then config['SWAGGER']
            # then default
            'info': self.config.get('info') or {
                'version': spec.get(
                    'version', self.config.get('version', '0.0.1')
                ),
                'title': spec.get(
                    'title', self.config.get('title', 'A swagger API')
                ),
                'description': spec.get(
                    'description', self.config.get('description',
                                                   'powered by Flasgger')
                ),
                'termsOfService': spec.get(
                    'termsOfService', self.config.get('termsOfService',
                                                      '/tos')
                ),
            },
            'paths': self.config.get('paths') or defaultdict(dict),
            'definitions': self.config.get('definitions') or defaultdict(dict)
        }

        openapi_version = self.config.get('openapi')
        if openapi_version:
            data['openapi'] = openapi_version
        else:
            data['swagger'] = self.config.get('swagger') or self.config.get(
                'swagger_version', '2.0'
            )

        # Support extension properties in the top level config
        top_level_extension_options = get_vendor_extension_fields(self.config)
        if top_level_extension_options:
            data.update(top_level_extension_options)

        # if True schema ids will be prefixed by function_method_{id}
        # for backwards compatibility with <= 0.5.14
        prefix_ids = self.config.get('prefix_ids')

        if self.config.get('host'):
            data['host'] = self.config.get('host')
        if self.config.get('basePath'):
            data['basePath'] = self.config.get('basePath')
        if self.config.get('schemes'):
            data['schemes'] = self.config.get('schemes')
        if self.config.get('securityDefinitions'):
            data['securityDefinitions'] = self.config.get(
                'securityDefinitions'
            )

        def is_openapi3():
            """
            returns True if `openapi_version` is 3.
            """

            return openapi_version and openapi_version.split('.')[0] == '3'

        if is_openapi3():
            # enable oas3 fields when openapi_version is 3.*.*
            optional_oas3_fields = self.config.get(
                'optional_oas3_fields') or OPTIONAL_OAS3_FIELDS
            for key in optional_oas3_fields:
                if self.config.get(key):
                    data[key] = self.config.get(key)

        # set defaults from template
        if self.template is not None:
            data.update(self.template)

        paths = data['paths']
        definitions = data['definitions']
        ignore_verbs = set(
            self.config.get('ignore_verbs', ('HEAD', 'OPTIONS'))
        )

        # technically only responses is non-optional
        optional_fields = self.config.get('optional_fields') or OPTIONAL_FIELDS

        for name, def_model in self.get_def_models(
                spec.get('definition_filter')).items():
            description, swag = parse_definition_docstring(
                def_model, self.sanitizer)
            if name and swag:
                if description:
                    swag.update({'description': description})
                definitions[name].update(swag)

        specs = self.get_specs(
            self.get_url_mappings(spec.get('rule_filter')), ignore_verbs,
            optional_fields, self.sanitizer,
            doc_dir=self.config.get('doc_dir'))

        http_methods = ['get', 'post', 'put', 'delete']
        for rule, verbs in specs:
            operations = dict()
            for verb, swag in verbs:
                update_dict = swag.get('definitions', {})
                if type(update_dict) == list and type(update_dict[0]) == dict:
                    # pop, assert single element
                    update_dict, = update_dict
                definitions.update(update_dict)
                defs = []  # swag.get('definitions', [])
                defs += extract_definitions(
                    defs, endpoint=rule.endpoint, verb=verb,
                    prefix_ids=prefix_ids
                )

                params = swag.get('parameters', [])
                if verb in swag.keys():
                    verb_swag = swag.get(verb)
                    if len(params) == 0 and verb.lower() in http_methods:
                        params = verb_swag.get('parameters', [])

                defs += extract_definitions(params,
                                            endpoint=rule.endpoint,
                                            verb=verb,
                                            prefix_ids=prefix_ids)

                request_body = swag.get('requestBody')
                if request_body:
                    content = request_body.get("content", {})
                    extract_definitions(
                        list(content.values()),
                        endpoint=rule.endpoint,
                        verb=verb,
                        prefix_ids=prefix_ids
                    )

                callbacks = swag.get('callbacks', {})
                if callbacks:
                    callbacks = {
                        str(key): value
                        for key, value in callbacks.items()
                    }
                    extract_definitions(
                        list(callbacks.values()),
                        endpoint=rule.endpoint,
                        verb=verb,
                        prefix_ids=prefix_ids
                    )

                responses = None
                if 'responses' in swag:
                    responses = swag.get('responses', {})
                    responses = {
                        str(key): value
                        for key, value in responses.items()
                    }
                    if responses is not None:
                        defs = defs + extract_definitions(
                            responses.values(),
                            endpoint=rule.endpoint,
                            verb=verb,
                            prefix_ids=prefix_ids
                        )
                    for definition in defs:
                        if 'id' not in definition:
                            definitions.update(definition)
                            continue
                        def_id = definition.pop('id')
                        if def_id is not None:
                            definitions[def_id].update(definition)

                operation = {}
                if swag.get('summary'):
                    operation['summary'] = swag.get('summary')
                if swag.get('description'):
                    operation['description'] = swag.get('description')
                if request_body:
                    operation['requestBody'] = request_body
                if callbacks:
                    operation['callbacks'] = callbacks
                if responses:
                    operation['responses'] = responses
                # parameters - swagger ui dislikes empty parameter lists
                if len(params) > 0:
                    operation['parameters'] = params
                # other optionals
                for key in optional_fields:
                    if key in swag:
                        value = swag.get(key)
                        if key in ('produces', 'consumes'):
                            if not isinstance(value, (list, tuple)):
                                value = [value]

                        operation[key] = value
                operations[verb] = operation

            if len(operations):
                try:
                    # Add reverse proxy prefix to route
                    prefix = self.template['swaggerUiPrefix']
                except (KeyError, TypeError):
                    prefix = ''
                srule = '{0}{1}'.format(prefix, rule)

                try:
                    # handle basePath
                    base_path = self.template['basePath']

                    if base_path:
                        if base_path.endswith('/'):
                            base_path = base_path[:-1]
                        if base_path:
                            # suppress base_path from srule if needed.
                            # Otherwise we will get definitions twice...
                            if srule.startswith(base_path):
                                srule = srule[len(base_path):]
                except (KeyError, TypeError):
                    pass

                # old regex '(<(.*?\:)?(.*?)>)'
                for arg in re.findall('(<([^<>]*:)?([^<>]*)>)', srule):
                    srule = srule.replace(arg[0], '{%s}' % arg[2])

                for key, val in operations.items():
                    if srule not in paths:
                        paths[srule] = {}
                    if key in paths[srule]:
                        paths[srule][key].update(val)
                    else:
                        paths[srule][key] = val
        self.apispecs[endpoint] = data
        return data

    def get_specs(self, rules, ignore_verbs, optional_fields, sanitizer, doc_dir=None):
        """
        gets api specs for given url rules.

        :param list[pyrin.api.router.handlers.base.RouteBase] rules: list of url rules.
        :param set[str] ignore_verbs: list of ignored http method names.
        :param list[str] optional_fields: list of optional fields in docstrings.
        :param function sanitizer: a callable to be used as string sanitizer.
        :param str doc_dir: directory path of docs.

        :rtype: list[tuple]
        """

        specs = []
        current_app = application_services.get_current_app()
        for rule in rules:
            if not self._should_expose_rule(rule):
                continue

            endpoint = current_app.view_functions[rule.endpoint]
            methods = dict()
            is_mv = is_valid_method_view(endpoint)

            for verb in rule.methods.difference(ignore_verbs):
                if not is_mv and has_valid_dispatch_view_docs(endpoint):
                    endpoint.methods = endpoint.methods or ['GET']
                    if verb in endpoint.methods:
                        methods[verb.lower()] = endpoint
                elif getattr(endpoint, 'methods', None) is not None:
                    if verb in endpoint.methods:
                        verb = verb.lower()
                        if hasattr(endpoint.view_class, verb):
                            methods[verb] = getattr(endpoint.view_class, verb)
                else:
                    methods[verb.lower()] = endpoint

            verbs = []
            for verb, method in methods.items():

                klass = method.__dict__.get('view_class', None)
                if not is_mv and klass and hasattr(klass, 'verb'):
                    method = getattr(klass, 'verb', None)
                elif klass and hasattr(klass, 'dispatch_request'):
                    method = getattr(klass, 'dispatch_request', None)
                if method is None:  # for MethodView
                    method = getattr(klass, verb, None)

                if method is None:
                    if is_mv:  # #76 Empty MethodViews
                        continue
                    raise RuntimeError(
                        'Cannot detect view_func for rule {0}'.format(rule)
                    )

                swag = {}

                if getattr(method, 'specs_dict', None):
                    definition = {}
                    merge_specs(
                        swag,
                        convert_schemas(deepcopy(method.specs_dict), definition)
                    )
                    swag['definitions'] = definition

                view_class = getattr(endpoint, 'view_class', None)
                if view_class and issubclass(view_class, SwaggerView):
                    apispec_swag = {}
                    apispec_attrs = optional_fields + [
                        'parameters', 'definitions', 'responses',
                        'summary', 'description'
                    ]
                    for attr in apispec_attrs:
                        value = getattr(view_class, attr)
                        if value:
                            apispec_swag[attr] = value

                    apispec_definitions = apispec_swag.get('definitions', {})
                    swag.update(
                        convert_schemas(apispec_swag, apispec_definitions)
                    )
                    swag['definitions'] = apispec_definitions

                if doc_dir:
                    if view_class:
                        file_path = os.path.join(
                            doc_dir, endpoint.__name__, method.__name__ + '.yml')
                    else:
                        file_path = os.path.join(
                            doc_dir, endpoint.__name__ + '.yml')
                    if os.path.isfile(file_path):
                        func = method.__func__ \
                            if hasattr(method, '__func__') else method
                        setattr(func, 'swag_type', 'yml')
                        setattr(func, 'swag_path', file_path)

                doc_summary, doc_description, doc_swag = parse_docstring(
                    method, sanitizer, endpoint=rule.endpoint, verb=verb)

                if doc_swag is None:
                    doc_swag = {}

                merge_specs(swag, doc_swag)

                if doc_summary:
                    swag['summary'] = doc_summary

                if doc_description:
                    swag['description'] = doc_description

                self._fix_metadata(rule, verb, swag)
                verbs.append((verb, swag))

            if verbs:
                specs.append((rule, verbs))

        return specs



# -*- coding: utf-8 -*-
"""
admin page schema module.
"""

from pyrin.api.schema.structs import ResultSchema
from pyrin.admin.interface import AbstractAdminPage
from pyrin.admin.exceptions import InvalidAdminPageTypeError


class AdminSchema(ResultSchema):
    """
    admin schema class.
    """

    def __init__(self, admin, **options):
        """
        initializes an instance of AdminSchema.

        note that at least one of keyword arguments must be provided.

        :param AbstractAdminPage admin: related admin page instance.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided.
                                                      it will be used only for entity
                                                      conversion.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will be
                                                           ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and `
                                                                    PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :raises InvalidAdminPageTypeError: invalid admin page type error.
        :raises SecureBooleanIsRequiredError: secure boolean is required error.
        :raises InvalidStartIndexError: invalid start index error.
        """

        if not isinstance(admin, AbstractAdminPage):
            raise InvalidAdminPageTypeError('Input parameter [{admin}] is '
                                            'not an instance of [{base}].'
                                            .format(admin=admin, base=AbstractAdminPage))

        super().__init__(**options)

        self._admin = admin

    def get_computed_row_columns(self, row, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        note that the result dict should not contain any `BaseEntity` or
        `ROW_RESULT` values, otherwise a max recursion error may occur.

        :param ROW_RESULT row: the actual row result to be processed.

        :rtype: dict
        """

        result = {}
        for method in self._admin.method_names:
            result[method] = self._admin.call_method(method, row)

        return result



# -*- coding: utf-8 -*-
"""
validator auto exceptions.
"""

from pyrin.core.exceptions import CoreException


class AutoValidatorException(CoreException):
    """
    auto validator exception.
    """
    pass


class InvalidAutoValidatorHookTypeError(AutoValidatorException):
    """
    invalid auto validator hook type error.
    """
    pass
